# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

# arv-copy [--recursive] [--no-recursive] object-uuid
#
# Copies an object from Arvados instance src to instance dst.
#
# By default, arv-copy recursively copies any dependent objects
# necessary to make the object functional in the new instance
# (e.g. for a workflow, arv-copy copies the workflow,
# input collections, and docker images). If
# --no-recursive is given, arv-copy copies only the single record
# identified by object-uuid.
#
# The user must have files $HOME/.config/arvados/{src}.conf and
# $HOME/.config/arvados/{dst}.conf with valid login credentials for
# instances src and dst.  If either of these files is not found,
# arv-copy will issue an error.

from __future__ import division
from future import standard_library
from future.utils import listvalues
standard_library.install_aliases()
from past.builtins import basestring
from builtins import object
import argparse
import contextlib
import getpass
import os
import re
import shutil
import sys
import logging
import tempfile
import urllib.parse
import io

import arvados
import arvados.config
import arvados.keep
import arvados.util
import arvados.commands._util as arv_cmd
import arvados.commands.keepdocker
import ruamel.yaml as yaml

from arvados.api import OrderedJsonModel
from arvados._version import __version__

COMMIT_HASH_RE = re.compile(r'^[0-9a-f]{1,40}$')

logger = logging.getLogger('arvados.arv-copy')

# local_repo_dir records which git repositories from the Arvados source
# instance have been checked out locally during this run, and to which
# directories.
# e.g. if repository 'twp' from src_arv has been cloned into
# /tmp/gitfHkV9lu44A then local_repo_dir['twp'] = '/tmp/gitfHkV9lu44A'
#
local_repo_dir = {}

# List of collections that have been copied in this session, and their
# destination collection UUIDs.
collections_copied = {}

# Set of (repository, script_version) two-tuples of commits copied in git.
scripts_copied = set()

# The owner_uuid of the object being copied
src_owner_uuid = None

def main():
    copy_opts = argparse.ArgumentParser(add_help=False)

    copy_opts.add_argument(
        '--version', action='version', version="%s %s" % (sys.argv[0], __version__),
        help='Print version and exit.')
    copy_opts.add_argument(
        '-v', '--verbose', dest='verbose', action='store_true',
        help='Verbose output.')
    copy_opts.add_argument(
        '--progress', dest='progress', action='store_true',
        help='Report progress on copying collections. (default)')
    copy_opts.add_argument(
        '--no-progress', dest='progress', action='store_false',
        help='Do not report progress on copying collections.')
    copy_opts.add_argument(
        '-f', '--force', dest='force', action='store_true',
        help='Perform copy even if the object appears to exist at the remote destination.')
    copy_opts.add_argument(
        '--src', dest='source_arvados',
        help='The cluster id of the source Arvados instance. May be either a pathname to a config file, or (for example) "foo" as shorthand for $HOME/.config/arvados/foo.conf.  If not provided, will be inferred from the UUID of the object being copied.')
    copy_opts.add_argument(
        '--dst', dest='destination_arvados',
        help='The name of the destination Arvados instance (required). May be either a pathname to a config file, or (for example) "foo" as shorthand for $HOME/.config/arvados/foo.conf.  If not provided, will use ARVADOS_API_HOST from environment.')
    copy_opts.add_argument(
        '--recursive', dest='recursive', action='store_true',
        help='Recursively copy any dependencies for this object, and subprojects. (default)')
    copy_opts.add_argument(
        '--no-recursive', dest='recursive', action='store_false',
        help='Do not copy any dependencies or subprojects.')
    copy_opts.add_argument(
        '--project-uuid', dest='project_uuid',
        help='The UUID of the project at the destination to which the collection or workflow should be copied.')
    copy_opts.add_argument(
        '--storage-classes', dest='storage_classes',
        help='Comma separated list of storage classes to be used when saving data to the destinaton Arvados instance.')

    copy_opts.add_argument(
        'object_uuid',
        help='The UUID of the object to be copied.')
    copy_opts.set_defaults(progress=True)
    copy_opts.set_defaults(recursive=True)

    parser = argparse.ArgumentParser(
        description='Copy a workflow, collection or project from one Arvados instance to another.  On success, the uuid of the copied object is printed to stdout.',
        parents=[copy_opts, arv_cmd.retry_opt])
    args = parser.parse_args()

    if args.storage_classes:
        args.storage_classes = [x for x in args.storage_classes.strip().replace(' ', '').split(',') if x]

    if args.verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if not args.source_arvados:
        args.source_arvados = args.object_uuid[:5]

    # Create API clients for the source and destination instances
    src_arv = api_for_instance(args.source_arvados)
    dst_arv = api_for_instance(args.destination_arvados)

    if not args.project_uuid:
        args.project_uuid = dst_arv.users().current().execute(num_retries=args.retries)["uuid"]

    # Identify the kind of object we have been given, and begin copying.
    t = uuid_type(src_arv, args.object_uuid)
    if t == 'Collection':
        set_src_owner_uuid(src_arv.collections(), args.object_uuid, args)
        result = copy_collection(args.object_uuid,
                                 src_arv, dst_arv,
                                 args)
    elif t == 'Workflow':
        set_src_owner_uuid(src_arv.workflows(), args.object_uuid, args)
        result = copy_workflow(args.object_uuid, src_arv, dst_arv, args)
    elif t == 'Group':
        set_src_owner_uuid(src_arv.groups(), args.object_uuid, args)
        result = copy_project(args.object_uuid, src_arv, dst_arv, args.project_uuid, args)
    else:
        abort("cannot copy object {} of type {}".format(args.object_uuid, t))

    # Clean up any outstanding temp git repositories.
    for d in listvalues(local_repo_dir):
        shutil.rmtree(d, ignore_errors=True)

    # If no exception was thrown and the response does not have an
    # error_token field, presume success
    if 'error_token' in result or 'uuid' not in result:
        logger.error("API server returned an error result: {}".format(result))
        exit(1)

    print(result['uuid'])

    if result.get('partial_error'):
        logger.warning("Warning: created copy with uuid {} but failed to copy some items: {}".format(result['uuid'], result['partial_error']))
        exit(1)

    logger.info("Success: created copy with uuid {}".format(result['uuid']))
    exit(0)

def set_src_owner_uuid(resource, uuid, args):
    global src_owner_uuid
    c = resource.get(uuid=uuid).execute(num_retries=args.retries)
    src_owner_uuid = c.get("owner_uuid")

# api_for_instance(instance_name)
#
#     Creates an API client for the Arvados instance identified by
#     instance_name.
#
#     If instance_name contains a slash, it is presumed to be a path
#     (either local or absolute) to a file with Arvados configuration
#     settings.
#
#     Otherwise, it is presumed to be the name of a file in
#     $HOME/.config/arvados/instance_name.conf
#
def api_for_instance(instance_name):
    if not instance_name:
        # Use environment
        return arvados.api('v1', model=OrderedJsonModel())

    if '/' in instance_name:
        config_file = instance_name
    else:
        config_file = os.path.join(os.environ['HOME'], '.config', 'arvados', "{}.conf".format(instance_name))

    try:
        cfg = arvados.config.load(config_file)
    except (IOError, OSError) as e:
        abort(("Could not open config file {}: {}\n" +
               "You must make sure that your configuration tokens\n" +
               "for Arvados instance {} are in {} and that this\n" +
               "file is readable.").format(
                   config_file, e, instance_name, config_file))

    if 'ARVADOS_API_HOST' in cfg and 'ARVADOS_API_TOKEN' in cfg:
        api_is_insecure = (
            cfg.get('ARVADOS_API_HOST_INSECURE', '').lower() in set(
                ['1', 't', 'true', 'y', 'yes']))
        client = arvados.api('v1',
                             host=cfg['ARVADOS_API_HOST'],
                             token=cfg['ARVADOS_API_TOKEN'],
                             insecure=api_is_insecure,
                             model=OrderedJsonModel())
    else:
        abort('need ARVADOS_API_HOST and ARVADOS_API_TOKEN for {}'.format(instance_name))
    return client

# Check if git is available
def check_git_availability():
    try:
        arvados.util.run_command(['git', '--help'])
    except Exception:
        abort('git command is not available. Please ensure git is installed.')


def filter_iter(arg):
    """Iterate a filter string-or-list.

    Pass in a filter field that can either be a string or list.
    This will iterate elements as if the field had been written as a list.
    """
    if isinstance(arg, basestring):
        return iter((arg,))
    else:
        return iter(arg)

def migrate_repository_filter(repo_filter, src_repository, dst_repository):
    """Update a single repository filter in-place for the destination.

    If the filter checks that the repository is src_repository, it is
    updated to check that the repository is dst_repository.  If it does
    anything else, this function raises ValueError.
    """
    if src_repository is None:
        raise ValueError("component does not specify a source repository")
    elif dst_repository is None:
        raise ValueError("no destination repository specified to update repository filter")
    elif repo_filter[1:] == ['=', src_repository]:
        repo_filter[2] = dst_repository
    elif repo_filter[1:] == ['in', [src_repository]]:
        repo_filter[2] = [dst_repository]
    else:
        raise ValueError("repository filter is not a simple source match")

def migrate_script_version_filter(version_filter):
    """Update a single script_version filter in-place for the destination.

    Currently this function checks that all the filter operands are Git
    commit hashes.  If they're not, it raises ValueError to indicate that
    the filter is not portable.  It could be extended to make other
    transformations in the future.
    """
    if not all(COMMIT_HASH_RE.match(v) for v in filter_iter(version_filter[2])):
        raise ValueError("script_version filter is not limited to commit hashes")

def attr_filtered(filter_, *attr_names):
    """Return True if filter_ applies to any of attr_names, else False."""
    return any((name == 'any') or (name in attr_names)
               for name in filter_iter(filter_[0]))

@contextlib.contextmanager
def exception_handler(handler, *exc_types):
    """If any exc_types are raised in the block, call handler on the exception."""
    try:
        yield
    except exc_types as error:
        handler(error)


# copy_workflow(wf_uuid, src, dst, args)
#
#    Copies a workflow identified by wf_uuid from src to dst.
#
#    If args.recursive is True, also copy any collections
#      referenced in the workflow definition yaml.
#
#    The owner_uuid of the new workflow is set to any given
#      project_uuid or the user who copied the template.
#
#    Returns the copied workflow object.
#
def copy_workflow(wf_uuid, src, dst, args):
    # fetch the workflow from the source instance
    wf = src.workflows().get(uuid=wf_uuid).execute(num_retries=args.retries)

    if not wf["definition"]:
        logger.warning("Workflow object {} has an empty or null definition, it won't do anything.".format(wf_uuid))

    # copy collections and docker images
    if args.recursive and wf["definition"]:
        wf_def = yaml.safe_load(wf["definition"])
        if wf_def is not None:
            locations = []
            docker_images = {}
            graph = wf_def.get('$graph', None)
            if graph is not None:
                workflow_collections(graph, locations, docker_images)
            else:
                workflow_collections(wf_def, locations, docker_images)

            if locations:
                copy_collections(locations, src, dst, args)

            for image in docker_images:
                copy_docker_image(image, docker_images[image], src, dst, args)

    # copy the workflow itself
    del wf['uuid']
    wf['owner_uuid'] = args.project_uuid

    existing = dst.workflows().list(filters=[["owner_uuid", "=", args.project_uuid],
                                             ["name", "=", wf["name"]]]).execute()
    if len(existing["items"]) == 0:
        return dst.workflows().create(body=wf).execute(num_retries=args.retries)
    else:
        return dst.workflows().update(uuid=existing["items"][0]["uuid"], body=wf).execute(num_retries=args.retries)


def workflow_collections(obj, locations, docker_images):
    if isinstance(obj, dict):
        loc = obj.get('location', None)
        if loc is not None:
            if loc.startswith("keep:"):
                locations.append(loc[5:])

        docker_image = obj.get('dockerImageId', None) or obj.get('dockerPull', None) or obj.get('acrContainerImage', None)
        if docker_image is not None:
            ds = docker_image.split(":", 1)
            tag = ds[1] if len(ds)==2 else 'latest'
            docker_images[ds[0]] = tag

        for x in obj:
            workflow_collections(obj[x], locations, docker_images)
    elif isinstance(obj, list):
        for x in obj:
            workflow_collections(x, locations, docker_images)

# copy_collections(obj, src, dst, args)
#
#    Recursively copies all collections referenced by 'obj' from src
#    to dst.  obj may be a dict or a list, in which case we run
#    copy_collections on every value it contains. If it is a string,
#    search it for any substring that matches a collection hash or uuid
#    (this will find hidden references to collections like
#      "input0": "$(file 3229739b505d2b878b62aed09895a55a+142/HWI-ST1027_129_D0THKACXX.1_1.fastq)")
#
#    Returns a copy of obj with any old collection uuids replaced by
#    the new ones.
#
def copy_collections(obj, src, dst, args):

    def copy_collection_fn(collection_match):
        """Helper function for regex substitution: copies a single collection,
        identified by the collection_match MatchObject, to the
        destination.  Returns the destination collection uuid (or the
        portable data hash if that's what src_id is).

        """
        src_id = collection_match.group(0)
        if src_id not in collections_copied:
            dst_col = copy_collection(src_id, src, dst, args)
            if src_id in [dst_col['uuid'], dst_col['portable_data_hash']]:
                collections_copied[src_id] = src_id
            else:
                collections_copied[src_id] = dst_col['uuid']
        return collections_copied[src_id]

    if isinstance(obj, basestring):
        # Copy any collections identified in this string to dst, replacing
        # them with the dst uuids as necessary.
        obj = arvados.util.portable_data_hash_pattern.sub(copy_collection_fn, obj)
        obj = arvados.util.collection_uuid_pattern.sub(copy_collection_fn, obj)
        return obj
    elif isinstance(obj, dict):
        return type(obj)((v, copy_collections(obj[v], src, dst, args))
                         for v in obj)
    elif isinstance(obj, list):
        return type(obj)(copy_collections(v, src, dst, args) for v in obj)
    return obj


def total_collection_size(manifest_text):
    """Return the total number of bytes in this collection (excluding
    duplicate blocks)."""

    total_bytes = 0
    locators_seen = {}
    for line in manifest_text.splitlines():
        words = line.split()
        for word in words[1:]:
            try:
                loc = arvados.KeepLocator(word)
            except ValueError:
                continue  # this word isn't a locator, skip it
            if loc.md5sum not in locators_seen:
                locators_seen[loc.md5sum] = True
                total_bytes += loc.size

    return total_bytes

def create_collection_from(c, src, dst, args):
    """Create a new collection record on dst, and copy Docker metadata if
    available."""

    collection_uuid = c['uuid']
    body = {}
    for d in ('description', 'manifest_text', 'name', 'portable_data_hash', 'properties'):
        body[d] = c[d]

    if not body["name"]:
        body['name'] = "copied from " + collection_uuid

    if args.storage_classes:
        body['storage_classes_desired'] = args.storage_classes

    body['owner_uuid'] = args.project_uuid

    dst_collection = dst.collections().create(body=body, ensure_unique_name=True).execute(num_retries=args.retries)

    # Create docker_image_repo+tag and docker_image_hash links
    # at the destination.
    for link_class in ("docker_image_repo+tag", "docker_image_hash"):
        docker_links = src.links().list(filters=[["head_uuid", "=", collection_uuid], ["link_class", "=", link_class]]).execute(num_retries=args.retries)['items']

        for src_link in docker_links:
            body = {key: src_link[key]
                    for key in ['link_class', 'name', 'properties']}
            body['head_uuid'] = dst_collection['uuid']
            body['owner_uuid'] = args.project_uuid

            lk = dst.links().create(body=body).execute(num_retries=args.retries)
            logger.debug('created dst link {}'.format(lk))

    return dst_collection

# copy_collection(obj_uuid, src, dst, args)
#
#    Copies the collection identified by obj_uuid from src to dst.
#    Returns the collection object created at dst.
#
#    If args.progress is True, produce a human-friendly progress
#    report.
#
#    If a collection with the desired portable_data_hash already
#    exists at dst, and args.force is False, copy_collection returns
#    the existing collection without copying any blocks.  Otherwise
#    (if no collection exists or if args.force is True)
#    copy_collection copies all of the collection data blocks from src
#    to dst.
#
#    For this application, it is critical to preserve the
#    collection's manifest hash, which is not guaranteed with the
#    arvados.CollectionReader and arvados.CollectionWriter classes.
#    Copying each block in the collection manually, followed by
#    the manifest block, ensures that the collection's manifest
#    hash will not change.
#
def copy_collection(obj_uuid, src, dst, args):
    if arvados.util.keep_locator_pattern.match(obj_uuid):
        # If the obj_uuid is a portable data hash, it might not be
        # uniquely identified with a particular collection.  As a
        # result, it is ambiguous as to what name to use for the copy.
        # Apply some heuristics to pick which collection to get the
        # name from.
        srccol = src.collections().list(
            filters=[['portable_data_hash', '=', obj_uuid]],
            order="created_at asc"
            ).execute(num_retries=args.retries)

        items = srccol.get("items")

        if not items:
            logger.warning("Could not find collection with portable data hash %s", obj_uuid)
            return

        c = None

        if len(items) == 1:
            # There's only one collection with the PDH, so use that.
            c = items[0]
        if not c:
            # See if there is a collection that's in the same project
            # as the root item (usually a workflow) being copied.
            for i in items:
                if i.get("owner_uuid") == src_owner_uuid and i.get("name"):
                    c = i
                    break
        if not c:
            # Didn't find any collections located in the same project, so
            # pick the oldest collection that has a name assigned to it.
            for i in items:
                if i.get("name"):
                    c = i
                    break
        if not c:
            # None of the collections have names (?!), so just pick the
            # first one.
            c = items[0]

        # list() doesn't return manifest text (and we don't want it to,
        # because we don't need the same maninfest text sent to us 50
        # times) so go and retrieve the collection object directly
        # which will include the manifest text.
        c = src.collections().get(uuid=c["uuid"]).execute(num_retries=args.retries)
    else:
        # Assume this is an actual collection uuid, so fetch it directly.
        c = src.collections().get(uuid=obj_uuid).execute(num_retries=args.retries)

    # If a collection with this hash already exists at the
    # destination, and 'force' is not true, just return that
    # collection.
    if not args.force:
        if 'portable_data_hash' in c:
            colhash = c['portable_data_hash']
        else:
            colhash = c['uuid']
        dstcol = dst.collections().list(
            filters=[['portable_data_hash', '=', colhash]]
        ).execute(num_retries=args.retries)
        if dstcol['items_available'] > 0:
            for d in dstcol['items']:
                if ((args.project_uuid == d['owner_uuid']) and
                    (c.get('name') == d['name']) and
                    (c['portable_data_hash'] == d['portable_data_hash'])):
                    return d
            c['manifest_text'] = dst.collections().get(
                uuid=dstcol['items'][0]['uuid']
            ).execute(num_retries=args.retries)['manifest_text']
            return create_collection_from(c, src, dst, args)

    # Fetch the collection's manifest.
    manifest = c['manifest_text']
    logger.debug("Copying collection %s with manifest: <%s>", obj_uuid, manifest)

    # Copy each block from src_keep to dst_keep.
    # Use the newly signed locators returned from dst_keep to build
    # a new manifest as we go.
    src_keep = arvados.keep.KeepClient(api_client=src, num_retries=args.retries)
    dst_keep = arvados.keep.KeepClient(api_client=dst, num_retries=args.retries)
    dst_manifest = io.StringIO()
    dst_locators = {}
    bytes_written = 0
    bytes_expected = total_collection_size(manifest)
    if args.progress:
        progress_writer = ProgressWriter(human_progress)
    else:
        progress_writer = None

    for line in manifest.splitlines():
        words = line.split()
        dst_manifest.write(words[0])
        for word in words[1:]:
            try:
                loc = arvados.KeepLocator(word)
            except ValueError:
                # If 'word' can't be parsed as a locator,
                # presume it's a filename.
                dst_manifest.write(' ')
                dst_manifest.write(word)
                continue
            blockhash = loc.md5sum
            # copy this block if we haven't seen it before
            # (otherwise, just reuse the existing dst_locator)
            if blockhash not in dst_locators:
                logger.debug("Copying block %s (%s bytes)", blockhash, loc.size)
                if progress_writer:
                    progress_writer.report(obj_uuid, bytes_written, bytes_expected)
                data = src_keep.get(word)
                dst_locator = dst_keep.put(data, classes=(args.storage_classes or []))
                dst_locators[blockhash] = dst_locator
                bytes_written += loc.size
            dst_manifest.write(' ')
            dst_manifest.write(dst_locators[blockhash])
        dst_manifest.write("\n")

    if progress_writer:
        progress_writer.report(obj_uuid, bytes_written, bytes_expected)
        progress_writer.finish()

    # Copy the manifest and save the collection.
    logger.debug('saving %s with manifest: <%s>', obj_uuid, dst_manifest.getvalue())

    c['manifest_text'] = dst_manifest.getvalue()
    return create_collection_from(c, src, dst, args)

def select_git_url(api, repo_name, retries, allow_insecure_http, allow_insecure_http_opt):
    r = api.repositories().list(
        filters=[['name', '=', repo_name]]).execute(num_retries=retries)
    if r['items_available'] != 1:
        raise Exception('cannot identify repo {}; {} repos found'
                        .format(repo_name, r['items_available']))

    https_url = [c for c in r['items'][0]["clone_urls"] if c.startswith("https:")]
    http_url = [c for c in r['items'][0]["clone_urls"] if c.startswith("http:")]
    other_url = [c for c in r['items'][0]["clone_urls"] if not c.startswith("http")]

    priority = https_url + other_url + http_url

    git_config = []
    git_url = None
    for url in priority:
        if url.startswith("http"):
            u = urllib.parse.urlsplit(url)
            baseurl = urllib.parse.urlunsplit((u.scheme, u.netloc, "", "", ""))
            git_config = ["-c", "credential.%s/.username=none" % baseurl,
                          "-c", "credential.%s/.helper=!cred(){ cat >/dev/null; if [ \"$1\" = get ]; then echo password=$ARVADOS_API_TOKEN; fi; };cred" % baseurl]
        else:
            git_config = []

        try:
            logger.debug("trying %s", url)
            arvados.util.run_command(["git"] + git_config + ["ls-remote", url],
                                      env={"HOME": os.environ["HOME"],
                                           "ARVADOS_API_TOKEN": api.api_token,
                                           "GIT_ASKPASS": "/bin/false"})
        except arvados.errors.CommandFailedError:
            pass
        else:
            git_url = url
            break

    if not git_url:
        raise Exception('Cannot access git repository, tried {}'
                        .format(priority))

    if git_url.startswith("http:"):
        if allow_insecure_http:
            logger.warning("Using insecure git url %s but will allow this because %s", git_url, allow_insecure_http_opt)
        else:
            raise Exception("Refusing to use insecure git url %s, use %s if you really want this." % (git_url, allow_insecure_http_opt))

    return (git_url, git_config)


def copy_docker_image(docker_image, docker_image_tag, src, dst, args):
    """Copy the docker image identified by docker_image and
    docker_image_tag from src to dst. Create appropriate
    docker_image_repo+tag and docker_image_hash links at dst.

    """

    logger.debug('copying docker image {}:{}'.format(docker_image, docker_image_tag))

    # Find the link identifying this docker image.
    docker_image_list = arvados.commands.keepdocker.list_images_in_arv(
        src, args.retries, docker_image, docker_image_tag)
    if docker_image_list:
        image_uuid, image_info = docker_image_list[0]
        logger.debug('copying collection {} {}'.format(image_uuid, image_info))

        # Copy the collection it refers to.
        dst_image_col = copy_collection(image_uuid, src, dst, args)
    elif arvados.util.keep_locator_pattern.match(docker_image):
        dst_image_col = copy_collection(docker_image, src, dst, args)
    else:
        logger.warning('Could not find docker image {}:{}'.format(docker_image, docker_image_tag))

def copy_project(obj_uuid, src, dst, owner_uuid, args):

    src_project_record = src.groups().get(uuid=obj_uuid).execute(num_retries=args.retries)

    # Create/update the destination project
    existing = dst.groups().list(filters=[["owner_uuid", "=", owner_uuid],
                                          ["name", "=", src_project_record["name"]]]).execute(num_retries=args.retries)
    if len(existing["items"]) == 0:
        project_record = dst.groups().create(body={"group": {"group_class": "project",
                                                             "owner_uuid": owner_uuid,
                                                             "name": src_project_record["name"]}}).execute(num_retries=args.retries)
    else:
        project_record = existing["items"][0]

    dst.groups().update(uuid=project_record["uuid"],
                        body={"group": {
                            "description": src_project_record["description"]}}).execute(num_retries=args.retries)

    args.project_uuid = project_record["uuid"]

    logger.debug('Copying %s to %s', obj_uuid, project_record["uuid"])


    partial_error = ""

    # Copy collections
    try:
        copy_collections([col["uuid"] for col in arvados.util.list_all(src.collections().list, filters=[["owner_uuid", "=", obj_uuid]])],
                         src, dst, args)
    except Exception as e:
        partial_error += "\n" + str(e)

    # Copy workflows
    for w in arvados.util.list_all(src.workflows().list, filters=[["owner_uuid", "=", obj_uuid]]):
        try:
            copy_workflow(w["uuid"], src, dst, args)
        except Exception as e:
            partial_error += "\n" + "Error while copying %s: %s" % (w["uuid"], e)

    if args.recursive:
        for g in arvados.util.list_all(src.groups().list, filters=[["owner_uuid", "=", obj_uuid]]):
            try:
                copy_project(g["uuid"], src, dst, project_record["uuid"], args)
            except Exception as e:
                partial_error += "\n" + "Error while copying %s: %s" % (g["uuid"], e)

    project_record["partial_error"] = partial_error

    return project_record

# git_rev_parse(rev, repo)
#
#    Returns the 40-character commit hash corresponding to 'rev' in
#    git repository 'repo' (which must be the path of a local git
#    repository)
#
def git_rev_parse(rev, repo):
    gitout, giterr = arvados.util.run_command(
        ['git', 'rev-parse', rev], cwd=repo)
    return gitout.strip()

# uuid_type(api, object_uuid)
#
#    Returns the name of the class that object_uuid belongs to, based on
#    the second field of the uuid.  This function consults the api's
#    schema to identify the object class.
#
#    It returns a string such as 'Collection', 'Workflow', etc.
#
#    Special case: if handed a Keep locator hash, return 'Collection'.
#
def uuid_type(api, object_uuid):
    if re.match(arvados.util.keep_locator_pattern, object_uuid):
        return 'Collection'
    p = object_uuid.split('-')
    if len(p) == 3:
        type_prefix = p[1]
        for k in api._schema.schemas:
            obj_class = api._schema.schemas[k].get('uuidPrefix', None)
            if type_prefix == obj_class:
                return k
    return None

def abort(msg, code=1):
    logger.info("arv-copy: %s", msg)
    exit(code)


# Code for reporting on the progress of a collection upload.
# Stolen from arvados.commands.put.ArvPutCollectionWriter
# TODO(twp): figure out how to refactor into a shared library
# (may involve refactoring some arvados.commands.arv_copy.copy_collection
# code)

def machine_progress(obj_uuid, bytes_written, bytes_expected):
    return "{} {}: {} {} written {} total\n".format(
        sys.argv[0],
        os.getpid(),
        obj_uuid,
        bytes_written,
        -1 if (bytes_expected is None) else bytes_expected)

def human_progress(obj_uuid, bytes_written, bytes_expected):
    if bytes_expected:
        return "\r{}: {}M / {}M {:.1%} ".format(
            obj_uuid,
            bytes_written >> 20, bytes_expected >> 20,
            float(bytes_written) / bytes_expected)
    else:
        return "\r{}: {} ".format(obj_uuid, bytes_written)

class ProgressWriter(object):
    _progress_func = None
    outfile = sys.stderr

    def __init__(self, progress_func):
        self._progress_func = progress_func

    def report(self, obj_uuid, bytes_written, bytes_expected):
        if self._progress_func is not None:
            self.outfile.write(
                self._progress_func(obj_uuid, bytes_written, bytes_expected))

    def finish(self):
        self.outfile.write("\n")

if __name__ == '__main__':
    main()



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
from builtins import object
import mock
import os
import unittest
import hashlib
from . import run_test_server
import json
import arvados
from . import arvados_testutil as tutil
from apiclient import http as apiclient_http


@tutil.skip_sleep
class ApiClientRetryTestMixin(object):

    TEST_UUID = 'zzzzz-zzzzz-zzzzzzzzzzzzzzz'
    TEST_LOCATOR = 'd41d8cd98f00b204e9800998ecf8427e+0'

    @classmethod
    def setUpClass(cls):
        run_test_server.run()

    def setUp(self):
        # Patch arvados.api() to return our mock API, so we can mock
        # its http requests.
        self.api_client = arvados.api('v1', cache=False)
        self.api_patch = mock.patch('arvados.api', return_value=self.api_client)
        self.api_patch.start()

    def tearDown(self):
        self.api_patch.stop()

    def run_method(self):
        raise NotImplementedError("test subclasses must define run_method")

    def test_immediate_success(self):
        with tutil.mock_api_responses(self.api_client, '{}', [200]):
            self.run_method()

    def test_immediate_failure(self):
        with tutil.mock_api_responses(self.api_client, '{}', [400]), self.assertRaises(self.DEFAULT_EXCEPTION):
            self.run_method()

    def test_retry_then_success(self):
        with tutil.mock_api_responses(self.api_client, '{}', [500, 200]):
            self.run_method()

    def test_error_after_default_retries_exhausted(self):
        with tutil.mock_api_responses(self.api_client, '{}', [500, 500, 500, 500, 500, 500, 200]), self.assertRaises(self.DEFAULT_EXCEPTION):
            self.run_method()

    def test_no_retry_after_immediate_success(self):
        with tutil.mock_api_responses(self.api_client, '{}', [200, 400]):
            self.run_method()


class CurrentJobTestCase(ApiClientRetryTestMixin, unittest.TestCase):

    DEFAULT_EXCEPTION = arvados.errors.ApiError

    def setUp(self):
        super(CurrentJobTestCase, self).setUp()
        os.environ['JOB_UUID'] = 'zzzzz-zzzzz-zzzzzzzzzzzzzzz'
        os.environ['JOB_WORK'] = '.'

    def tearDown(self):
        del os.environ['JOB_UUID']
        del os.environ['JOB_WORK']
        arvados._current_job = None
        super(CurrentJobTestCase, self).tearDown()

    def run_method(self):
        arvados.current_job()



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

import logging

from . import api

_logger = logging.getLogger('arvados.vocabulary')

def load_vocabulary(api_client=None):
    """Load the Arvados vocabulary from the API.
    """
    if api_client is None:
        api_client = api('v1')
    return Vocabulary(api_client.vocabulary())

class VocabularyError(Exception):
    """Base class for all vocabulary errors.
    """
    pass

class VocabularyKeyError(VocabularyError):
    pass

class VocabularyValueError(VocabularyError):
    pass

class Vocabulary(object):
    def __init__(self, voc_definition={}):
        self.strict_keys = voc_definition.get('strict_tags', False)
        self.key_aliases = {}

        for key_id, val in (voc_definition.get('tags') or {}).items():
            strict = val.get('strict', False)
            key_labels = [l['label'] for l in val.get('labels', [])]
            values = {}
            for v_id, v_val in (val.get('values') or {}).items():
                labels = [l['label'] for l in v_val.get('labels', [])]
                values[v_id] = VocabularyValue(v_id, labels)
            vk = VocabularyKey(key_id, key_labels, values, strict)
            self.key_aliases[key_id.lower()] = vk
            for alias in vk.aliases:
                self.key_aliases[alias.lower()] = vk

    def __getitem__(self, key):
        return self.key_aliases[key.lower()]

    def convert_to_identifiers(self, obj={}):
        """Translate key/value pairs to machine readable identifiers.
        """
        return self._convert_to_what(obj, 'identifier')

    def convert_to_labels(self, obj={}):
        """Translate key/value pairs to human readable labels.
        """
        return self._convert_to_what(obj, 'preferred_label')

    def _convert_to_what(self, obj={}, what=None):
        if not isinstance(obj, dict):
            raise ValueError("obj must be a dict")
        if what not in ['preferred_label', 'identifier']:
            raise ValueError("what attr must be 'preferred_label' or 'identifier'")
        r = {}
        for k, v in obj.items():
            # Key validation & lookup
            key_found = False
            if not isinstance(k, str):
                raise VocabularyKeyError("key '{}' must be a string".format(k))
            k_what, v_what = k, v
            try:
                k_what = getattr(self[k], what)
                key_found = True
            except KeyError:
                if self.strict_keys:
                    raise VocabularyKeyError("key '{}' not found in vocabulary".format(k))

            # Value validation & lookup
            if isinstance(v, list):
                v_what = []
                for x in v:
                    if not isinstance(x, str):
                        raise VocabularyValueError("value '{}' for key '{}' must be a string".format(x, k))
                    try:
                        v_what.append(getattr(self[k][x], what))
                    except KeyError:
                        if self[k].strict:
                            raise VocabularyValueError("value '{}' not found for key '{}'".format(x, k))
                        v_what.append(x)
            else:
                if not isinstance(v, str):
                    raise VocabularyValueError("{} value '{}' for key '{}' must be a string".format(type(v).__name__, v, k))
                try:
                    v_what = getattr(self[k][v], what)
                except KeyError:
                    if key_found and self[k].strict:
                        raise VocabularyValueError("value '{}' not found for key '{}'".format(v, k))

            r[k_what] = v_what
        return r

class VocabularyData(object):
    def __init__(self, identifier, aliases=[]):
        self.identifier = identifier
        self.aliases = aliases

    def __getattribute__(self, name):
        if name == 'preferred_label':
            return self.aliases[0]
        return super(VocabularyData, self).__getattribute__(name)

class VocabularyValue(VocabularyData):
    def __init__(self, identifier, aliases=[]):
        super(VocabularyValue, self).__init__(identifier, aliases)

class VocabularyKey(VocabularyData):
    def __init__(self, identifier, aliases=[], values={}, strict=False):
        super(VocabularyKey, self).__init__(identifier, aliases)
        self.strict = strict
        self.value_aliases = {}
        for v_id, v_val in values.items():
            self.value_aliases[v_id.lower()] = v_val
            for v_alias in v_val.aliases:
                self.value_aliases[v_alias.lower()] = v_val

    def __getitem__(self, key):
        return self.value_aliases[key.lower()]


# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

# config.py - configuration settings and global variables for Arvados clients
#
# Arvados configuration settings are taken from $HOME/.config/arvados.
# Environment variables override settings in the config file.

import os
import re

_settings = None
if os.environ.get('HOME') is not None:
    default_config_file = os.environ['HOME'] + '/.config/arvados/settings.conf'
else:
    default_config_file = ''

KEEP_BLOCK_SIZE = 2**26
EMPTY_BLOCK_LOCATOR = 'd41d8cd98f00b204e9800998ecf8427e+0'

def initialize(config_file=default_config_file):
    global _settings
    _settings = {}

    # load the specified config file if available
    try:
        _settings = load(config_file)
    except IOError:
        pass

    # override any settings with environment vars
    for var in os.environ:
        if var.startswith('ARVADOS_'):
            _settings[var] = os.environ[var]

def load(config_file):
    cfg = {}
    with open(config_file, "r") as f:
        for config_line in f:
            if re.match('^\s*$', config_line):
                continue
            if re.match('^\s*#', config_line):
                continue
            var, val = config_line.rstrip().split('=', 2)
            cfg[var] = val
    return cfg

def flag_is_true(key, d=None):
    if d is None:
        d = settings()
    return d.get(key, '').lower() in set(['1', 't', 'true', 'y', 'yes'])

def get(key, default_val=None):
    return settings().get(key, default_val)

def settings():
    if _settings is None:
        initialize()
    return _settings



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

import os
import subprocess
import unittest

import arvados
import arvados.util

class MkdirDashPTest(unittest.TestCase):
    def setUp(self):
        try:
            os.path.mkdir('./tmp')
        except:
            pass
    def tearDown(self):
        try:
            os.unlink('./tmp/bar')
            os.rmdir('./tmp/foo')
            os.rmdir('./tmp')
        except:
            pass
    def runTest(self):
        arvados.util.mkdir_dash_p('./tmp/foo')
        with open('./tmp/bar', 'wb') as f:
            f.write(b'bar')
        self.assertRaises(OSError, arvados.util.mkdir_dash_p, './tmp/bar')


class RunCommandTestCase(unittest.TestCase):
    def test_success(self):
        stdout, stderr = arvados.util.run_command(['echo', 'test'],
                                                  stderr=subprocess.PIPE)
        self.assertEqual("test\n".encode(), stdout)
        self.assertEqual("".encode(), stderr)

    def test_failure(self):
        with self.assertRaises(arvados.errors.CommandFailedError):
            arvados.util.run_command(['false'])

class KeysetTestHelper:
    def __init__(self, expect):
        self.n = 0
        self.expect = expect

    def fn(self, **kwargs):
        if self.expect[self.n][0] != kwargs:
            raise Exception("Didn't match %s != %s" % (self.expect[self.n][0], kwargs))
        return self

    def execute(self, num_retries):
        self.n += 1
        return self.expect[self.n-1][1]

class KeysetListAllTestCase(unittest.TestCase):
    def test_empty(self):
        ks = KeysetTestHelper([[
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": []},
            {"items": []}
        ]])

        ls = list(arvados.util.keyset_list_all(ks.fn))
        self.assertEqual(ls, [])

    def test_oneitem(self):
        ks = KeysetTestHelper([[
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": []},
            {"items": [{"created_at": "1", "uuid": "1"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", "=", "1"], ["uuid", ">", "1"]]},
            {"items": []}
        ],[
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", ">", "1"]]},
            {"items": []}
        ]])

        ls = list(arvados.util.keyset_list_all(ks.fn))
        self.assertEqual(ls, [{"created_at": "1", "uuid": "1"}])

    def test_onepage2(self):
        ks = KeysetTestHelper([[
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": []},
            {"items": [{"created_at": "1", "uuid": "1"}, {"created_at": "2", "uuid": "2"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", ">=", "2"], ["uuid", "!=", "2"]]},
            {"items": []}
        ]])

        ls = list(arvados.util.keyset_list_all(ks.fn))
        self.assertEqual(ls, [{"created_at": "1", "uuid": "1"}, {"created_at": "2", "uuid": "2"}])

    def test_onepage3(self):
        ks = KeysetTestHelper([[
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": []},
            {"items": [{"created_at": "1", "uuid": "1"}, {"created_at": "2", "uuid": "2"}, {"created_at": "3", "uuid": "3"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", ">=", "3"], ["uuid", "!=", "3"]]},
            {"items": []}
        ]])

        ls = list(arvados.util.keyset_list_all(ks.fn))
        self.assertEqual(ls, [{"created_at": "1", "uuid": "1"}, {"created_at": "2", "uuid": "2"}, {"created_at": "3", "uuid": "3"}])


    def test_twopage(self):
        ks = KeysetTestHelper([[
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": []},
            {"items": [{"created_at": "1", "uuid": "1"}, {"created_at": "2", "uuid": "2"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", ">=", "2"], ["uuid", "!=", "2"]]},
            {"items": [{"created_at": "3", "uuid": "3"}, {"created_at": "4", "uuid": "4"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", ">=", "4"], ["uuid", "!=", "4"]]},
            {"items": []}
        ]])

        ls = list(arvados.util.keyset_list_all(ks.fn))
        self.assertEqual(ls, [{"created_at": "1", "uuid": "1"},
                              {"created_at": "2", "uuid": "2"},
                              {"created_at": "3", "uuid": "3"},
                              {"created_at": "4", "uuid": "4"}
        ])

    def test_repeated_key(self):
        ks = KeysetTestHelper([[
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": []},
            {"items": [{"created_at": "1", "uuid": "1"}, {"created_at": "2", "uuid": "2"}, {"created_at": "2", "uuid": "3"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", ">=", "2"], ["uuid", "!=", "3"]]},
            {"items": [{"created_at": "2", "uuid": "2"}, {"created_at": "2", "uuid": "4"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", "=", "2"], ["uuid", ">", "4"]]},
            {"items": []}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", ">", "2"]]},
            {"items": [{"created_at": "3", "uuid": "5"}, {"created_at": "4", "uuid": "6"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", ">=", "4"], ["uuid", "!=", "6"]]},
            {"items": []}
        ],
        ])

        ls = list(arvados.util.keyset_list_all(ks.fn))
        self.assertEqual(ls, [{"created_at": "1", "uuid": "1"},
                              {"created_at": "2", "uuid": "2"},
                              {"created_at": "2", "uuid": "3"},
                              {"created_at": "2", "uuid": "4"},
                              {"created_at": "3", "uuid": "5"},
                              {"created_at": "4", "uuid": "6"}
        ])

    def test_onepage_withfilter(self):
        ks = KeysetTestHelper([[
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["foo", ">", "bar"]]},
            {"items": [{"created_at": "1", "uuid": "1"}, {"created_at": "2", "uuid": "2"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at asc", "uuid asc"], "filters": [["created_at", ">=", "2"], ["uuid", "!=", "2"], ["foo", ">", "bar"]]},
            {"items": []}
        ]])

        ls = list(arvados.util.keyset_list_all(ks.fn, filters=[["foo", ">", "bar"]]))
        self.assertEqual(ls, [{"created_at": "1", "uuid": "1"}, {"created_at": "2", "uuid": "2"}])


    def test_onepage_desc(self):
        ks = KeysetTestHelper([[
            {"limit": 1000, "count": "none", "order": ["created_at desc", "uuid desc"], "filters": []},
            {"items": [{"created_at": "2", "uuid": "2"}, {"created_at": "1", "uuid": "1"}]}
        ], [
            {"limit": 1000, "count": "none", "order": ["created_at desc", "uuid desc"], "filters": [["created_at", "<=", "1"], ["uuid", "!=", "1"]]},
            {"items": []}
        ]])

        ls = list(arvados.util.keyset_list_all(ks.fn, ascending=False))
        self.assertEqual(ls, [{"created_at": "2", "uuid": "2"}, {"created_at": "1", "uuid": "1"}])



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import zip
from builtins import range
from builtins import object
import itertools
import unittest

import arvados.errors as arv_error
import arvados.retry as arv_retry
import mock

class RetryLoopTestMixin(object):
    @staticmethod
    def loop_success(result):
        # During the tests, we use integers that look like HTTP status
        # codes as loop results.  Then we define simplified HTTP
        # heuristics here to decide whether the result is success (True),
        # permanent failure (False), or temporary failure (None).
        if result < 400:
            return True
        elif result < 500:
            return False
        else:
            return None

    def run_loop(self, num_retries, *results, **kwargs):
        responses = itertools.chain(results, itertools.repeat(None))
        retrier = arv_retry.RetryLoop(num_retries, self.loop_success,
                                      **kwargs)
        for tries_left, response in zip(retrier, responses):
            retrier.save_result(response)
        return retrier

    def check_result(self, retrier, expect_success, last_code):
        self.assertIs(retrier.success(), expect_success,
                      "loop success flag is incorrect")
        self.assertEqual(last_code, retrier.last_result())


class RetryLoopTestCase(unittest.TestCase, RetryLoopTestMixin):
    def test_zero_retries_and_success(self):
        retrier = self.run_loop(0, 200)
        self.check_result(retrier, True, 200)

    def test_zero_retries_and_tempfail(self):
        retrier = self.run_loop(0, 500, 501)
        self.check_result(retrier, None, 500)

    def test_zero_retries_and_permfail(self):
        retrier = self.run_loop(0, 400, 201)
        self.check_result(retrier, False, 400)

    def test_one_retry_with_immediate_success(self):
        retrier = self.run_loop(1, 200, 201)
        self.check_result(retrier, True, 200)

    def test_one_retry_with_delayed_success(self):
        retrier = self.run_loop(1, 500, 201)
        self.check_result(retrier, True, 201)

    def test_one_retry_with_no_success(self):
        retrier = self.run_loop(1, 500, 501, 502)
        self.check_result(retrier, None, 501)

    def test_one_retry_but_permfail(self):
        retrier = self.run_loop(1, 400, 201)
        self.check_result(retrier, False, 400)

    def test_two_retries_with_immediate_success(self):
        retrier = self.run_loop(2, 200, 201, 202)
        self.check_result(retrier, True, 200)

    def test_two_retries_with_success_after_one(self):
        retrier = self.run_loop(2, 500, 201, 502)
        self.check_result(retrier, True, 201)

    def test_two_retries_with_success_after_two(self):
        retrier = self.run_loop(2, 500, 501, 202, 503)
        self.check_result(retrier, True, 202)

    def test_two_retries_with_no_success(self):
        retrier = self.run_loop(2, 500, 501, 502, 503)
        self.check_result(retrier, None, 502)

    def test_two_retries_with_permfail(self):
        retrier = self.run_loop(2, 500, 401, 202)
        self.check_result(retrier, False, 401)

    def test_save_result_before_start_is_error(self):
        retrier = arv_retry.RetryLoop(0)
        self.assertRaises(arv_error.AssertionError, retrier.save_result, 1)

    def test_save_result_after_end_is_error(self):
        retrier = arv_retry.RetryLoop(0)
        for count in retrier:
            pass
        self.assertRaises(arv_error.AssertionError, retrier.save_result, 1)


@mock.patch('time.time', side_effect=itertools.count())
@mock.patch('time.sleep')
class RetryLoopBackoffTestCase(unittest.TestCase, RetryLoopTestMixin):
    def run_loop(self, num_retries, *results, **kwargs):
        kwargs.setdefault('backoff_start', 8)
        return super(RetryLoopBackoffTestCase, self).run_loop(
            num_retries, *results, **kwargs)

    def check_backoff(self, sleep_mock, sleep_count, multiplier=1):
        # Figure out how much time we actually spent sleeping.
        sleep_times = [arglist[0][0] for arglist in sleep_mock.call_args_list
                       if arglist[0][0] > 0]
        self.assertEqual(sleep_count, len(sleep_times),
                         "loop did not back off correctly")
        last_wait = 0
        for this_wait in sleep_times:
            self.assertGreater(this_wait, last_wait * multiplier,
                               "loop did not grow backoff times correctly")
            last_wait = this_wait

    def test_no_backoff_with_no_retries(self, sleep_mock, time_mock):
        self.run_loop(0, 500, 201)
        self.check_backoff(sleep_mock, 0)

    def test_no_backoff_after_success(self, sleep_mock, time_mock):
        self.run_loop(1, 200, 501)
        self.check_backoff(sleep_mock, 0)

    def test_no_backoff_after_permfail(self, sleep_mock, time_mock):
        self.run_loop(1, 400, 201)
        self.check_backoff(sleep_mock, 0)

    def test_backoff_before_success(self, sleep_mock, time_mock):
        self.run_loop(5, 500, 501, 502, 203, 504)
        self.check_backoff(sleep_mock, 3)

    def test_backoff_before_permfail(self, sleep_mock, time_mock):
        self.run_loop(5, 500, 501, 502, 403, 504)
        self.check_backoff(sleep_mock, 3)

    def test_backoff_all_tempfail(self, sleep_mock, time_mock):
        self.run_loop(3, 500, 501, 502, 503, 504)
        self.check_backoff(sleep_mock, 3)

    def test_backoff_multiplier(self, sleep_mock, time_mock):
        self.run_loop(5, 500, 501, 502, 503, 504, 505,
                      backoff_start=5, backoff_growth=10, max_wait=1000000000)
        self.check_backoff(sleep_mock, 5, 9)


class CheckHTTPResponseSuccessTestCase(unittest.TestCase):
    def results_map(self, *codes):
        for code in codes:
            yield code, arv_retry.check_http_response_success(code)

    def check(assert_name):
        def check_method(self, expected, *codes):
            assert_func = getattr(self, assert_name)
            for code, actual in self.results_map(*codes):
                assert_func(expected, actual,
                            "{} status flagged {}".format(code, actual))
                if assert_name != 'assertIs':
                    self.assertTrue(
                        actual is True or actual is False or actual is None,
                        "{} status returned {}".format(code, actual))
        return check_method

    check_is = check('assertIs')
    check_is_not = check('assertIsNot')

    def test_obvious_successes(self):
        self.check_is(True, *list(range(200, 207)))

    def test_obvious_stops(self):
        self.check_is(False, 424, 426, 428, 431,
                      *list(range(400, 408)) + list(range(410, 420)))

    def test_obvious_retries(self):
        self.check_is(None, 500, 502, 503, 504)

    def test_4xx_retries(self):
        self.check_is(None, 408, 409, 422, 423)

    def test_5xx_failures(self):
        self.check_is(False, 501, *list(range(505, 512)))

    def test_1xx_not_retried(self):
        self.check_is_not(None, 100, 101)

    def test_redirects_not_retried(self):
        self.check_is_not(None, *list(range(300, 309)))

    def test_wacky_code_retries(self):
        self.check_is(None, 0, 99, 600, -200)


class RetryMethodTestCase(unittest.TestCase):
    class Tester(object):
        def __init__(self):
            self.num_retries = 1

        @arv_retry.retry_method
        def check(self, a, num_retries=None, z=0):
            return (a, num_retries, z)


    def test_positional_arg_raises(self):
        # unsupported use -- make sure we raise rather than ignore
        with self.assertRaises(TypeError):
            self.assertEqual((3, 2, 0), self.Tester().check(3, 2))

    def test_keyword_arg_passed(self):
        self.assertEqual((4, 3, 0), self.Tester().check(num_retries=3, a=4))

    def test_not_specified(self):
        self.assertEqual((0, 1, 0), self.Tester().check(0))

    def test_not_specified_with_other_kwargs(self):
        self.assertEqual((1, 1, 1), self.Tester().check(1, z=1))

    def test_bad_call(self):
        with self.assertRaises(TypeError):
            self.Tester().check(num_retries=2)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import object
import json
import os

class TaskOutputDir(object):
    """Keep-backed directory for staging outputs of Crunch tasks.

    Example, in a crunch task whose output is a file called "out.txt"
    containing "42":

        import arvados
        import arvados.crunch
        import os

        out = arvados.crunch.TaskOutputDir()
        with open(os.path.join(out.path, 'out.txt'), 'w') as f:
            f.write('42')
        arvados.current_task().set_output(out.manifest_text())
    """
    def __init__(self):
        self.path = os.environ['TASK_KEEPMOUNT_TMP']

    def __str__(self):
        return self.path

    def manifest_text(self):
        snapshot = os.path.join(self.path, '.arvados#collection')
        return json.load(open(snapshot))['manifest_text']



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import str
import http.server
import hashlib
import os
import re
import socket
import socketserver
import sys
import threading
import time

from . import arvados_testutil as tutil

_debug = os.environ.get('ARVADOS_DEBUG', None)


class StubKeepServers(tutil.ApiClientMock):

    def setUp(self):
        super(StubKeepServers, self).setUp()
        sock = socket.socket()
        sock.bind(('0.0.0.0', 0))
        self.port = sock.getsockname()[1]
        sock.close()
        self.server = Server(('0.0.0.0', self.port), Handler)
        self.thread = threading.Thread(target=self.server.serve_forever)
        self.thread.daemon = True # Exit thread if main proc exits
        self.thread.start()
        self.api_client = self.mock_keep_services(
            count=1,
            service_host='localhost',
            service_port=self.port,
        )

    def tearDown(self):
        self.server.shutdown()
        super(StubKeepServers, self).tearDown()


class Server(socketserver.ThreadingMixIn, http.server.HTTPServer, object):

    allow_reuse_address = 1

    def __init__(self, *args, **kwargs):
        self.store = {}
        self.delays = {
            # before reading request headers
            'request': 0,
            # before reading request body
            'request_body': 0,
            # before setting response status and headers
            'response': 0,
            # before sending response body
            'response_body': 0,
            # before returning from handler (thus setting response EOF)
            'response_close': 0,
            # after writing over 1s worth of data at self.bandwidth
            'mid_write': 0,
            # after reading over 1s worth of data at self.bandwidth
            'mid_read': 0,
        }
        self.bandwidth = None
        super(Server, self).__init__(*args, **kwargs)

    def setdelays(self, **kwargs):
        """In future requests, induce delays at the given checkpoints."""
        for (k, v) in kwargs.items():
            self.delays.get(k)  # NameError if unknown key
            self.delays[k] = v

    def setbandwidth(self, bandwidth):
        """For future requests, set the maximum bandwidth (number of bytes per
        second) to operate at. If setbandwidth is never called, function at
        maximum bandwidth possible"""
        self.bandwidth = float(bandwidth)

    def _sleep_at_least(self, seconds):
        """Sleep for given time, even if signals are received."""
        wake = time.time() + seconds
        todo = seconds
        while todo > 0:
            time.sleep(todo)
            todo = wake - time.time()

    def _do_delay(self, k):
        self._sleep_at_least(self.delays[k])


class Handler(http.server.BaseHTTPRequestHandler, object):

    protocol_version = 'HTTP/1.1'

    def wfile_bandwidth_write(self, data_to_write):
        if self.server.bandwidth is None and self.server.delays['mid_write'] == 0:
            self.wfile.write(data_to_write)
        else:
            BYTES_PER_WRITE = int(self.server.bandwidth/4) or 32768
            outage_happened = False
            num_bytes = len(data_to_write)
            num_sent_bytes = 0
            target_time = time.time()
            while num_sent_bytes < num_bytes:
                if num_sent_bytes > self.server.bandwidth and not outage_happened:
                    self.server._do_delay('mid_write')
                    target_time += self.server.delays['mid_write']
                    outage_happened = True
                num_write_bytes = min(BYTES_PER_WRITE,
                    num_bytes - num_sent_bytes)
                self.wfile.write(data_to_write[
                    num_sent_bytes:num_sent_bytes+num_write_bytes])
                num_sent_bytes += num_write_bytes
                if self.server.bandwidth is not None:
                    target_time += num_write_bytes / self.server.bandwidth
                    self.server._sleep_at_least(target_time - time.time())
        return None

    def rfile_bandwidth_read(self, bytes_to_read):
        if self.server.bandwidth is None and self.server.delays['mid_read'] == 0:
            return self.rfile.read(bytes_to_read)
        else:
            BYTES_PER_READ = int(self.server.bandwidth/4) or 32768
            data = b''
            outage_happened = False
            bytes_read = 0
            target_time = time.time()
            while bytes_to_read > bytes_read:
                if bytes_read > self.server.bandwidth and not outage_happened:
                    self.server._do_delay('mid_read')
                    target_time += self.server.delays['mid_read']
                    outage_happened = True
                next_bytes_to_read = min(BYTES_PER_READ,
                    bytes_to_read - bytes_read)
                data += self.rfile.read(next_bytes_to_read)
                bytes_read += next_bytes_to_read
                if self.server.bandwidth is not None:
                    target_time += next_bytes_to_read / self.server.bandwidth
                    self.server._sleep_at_least(target_time - time.time())
        return data

    def finish(self, *args, **kwargs):
        try:
            return super(Handler, self).finish(*args, **kwargs)
        except Exception as err:
            if _debug:
                raise

    def handle(self, *args, **kwargs):
        try:
            return super(Handler, self).handle(*args, **kwargs)
        except:
            if _debug:
                raise

    def handle_one_request(self, *args, **kwargs):
        self._sent_continue = False
        self.server._do_delay('request')
        return super(Handler, self).handle_one_request(*args, **kwargs)

    def handle_expect_100(self):
        self.server._do_delay('request_body')
        self._sent_continue = True
        return super(Handler, self).handle_expect_100()

    def do_GET(self):
        self.server._do_delay('response')
        r = re.search(r'[0-9a-f]{32}', self.path)
        if not r:
            return self.send_response(422)
        datahash = r.group(0)
        if datahash not in self.server.store:
            return self.send_response(404)
        self.send_response(200)
        self.send_header('Connection', 'close')
        self.send_header('Content-type', 'application/octet-stream')
        self.end_headers()
        self.server._do_delay('response_body')
        self.wfile_bandwidth_write(self.server.store[datahash])
        self.server._do_delay('response_close')

    def do_HEAD(self):
        self.server._do_delay('response')
        r = re.search(r'[0-9a-f]{32}', self.path)
        if not r:
            return self.send_response(422)
        datahash = r.group(0)
        if datahash not in self.server.store:
            return self.send_response(404)
        self.send_response(200)
        self.send_header('Connection', 'close')
        self.send_header('Content-type', 'application/octet-stream')
        self.send_header('Content-length', str(len(self.server.store[datahash])))
        self.end_headers()
        self.server._do_delay('response_close')
        self.close_connection = True

    def do_PUT(self):
        if not self._sent_continue and self.headers.get('expect') == '100-continue':
            # The comments at https://bugs.python.org/issue1491
            # implies that Python 2.7 BaseHTTPRequestHandler was
            # patched to support 100 Continue, but reading the actual
            # code that ships in Debian it clearly is not, so we need
            # to send the response on the socket directly.
            self.server._do_delay('request_body')
            self.wfile.write("{} {} {}\r\n\r\n".format(
                self.protocol_version, 100, "Continue").encode())
        data = self.rfile_bandwidth_read(
            int(self.headers.get('content-length')))
        datahash = hashlib.md5(data).hexdigest()
        self.server.store[datahash] = data
        resp = '{}+{}\n'.format(datahash, len(data)).encode()
        self.server._do_delay('response')
        self.send_response(200)
        self.send_header('Connection', 'close')
        self.send_header('Content-type', 'text/plain')
        self.send_header('Content-length', len(resp))
        self.end_headers()
        self.server._do_delay('response_body')
        self.wfile_bandwidth_write(resp)
        self.server._do_delay('response_close')
        self.close_connection = True

    def log_request(self, *args, **kwargs):
        if _debug:
            super(Handler, self).log_request(*args, **kwargs)



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
import os
import sys
import tempfile
import unittest
import shutil
import arvados.api
from arvados.collection import Collection, CollectionReader

import arvados.commands.arv_copy as arv_copy
from . import arvados_testutil as tutil
from . import run_test_server

class ArvCopyVersionTestCase(run_test_server.TestCaseWithServers, tutil.VersionChecker):
    MAIN_SERVER = {}
    KEEP_SERVER = {}

    def run_copy(self, args):
        sys.argv = ['arv-copy'] + args
        return arv_copy.main()

    def test_unsupported_arg(self):
        with self.assertRaises(SystemExit):
            self.run_copy(['-x=unknown'])

    def test_version_argument(self):
        with tutil.redirected_streams(
                stdout=tutil.StringIO, stderr=tutil.StringIO) as (out, err):
            with self.assertRaises(SystemExit):
                self.run_copy(['--version'])
        self.assertVersionOutput(out, err)

    def test_copy_project(self):
        api = arvados.api()
        src_proj = api.groups().create(body={"group": {"name": "arv-copy project", "group_class": "project"}}).execute()["uuid"]

        c = Collection()
        with c.open('foo', 'wt') as f:
            f.write('foo')
        c.save_new("arv-copy foo collection", owner_uuid=src_proj)
        coll_record = api.collections().get(uuid=c.manifest_locator()).execute()
        assert coll_record['storage_classes_desired'] == ['default']

        dest_proj = api.groups().create(body={"group": {"name": "arv-copy dest project", "group_class": "project"}}).execute()["uuid"]

        tmphome = tempfile.mkdtemp()
        home_was = os.environ['HOME']
        os.environ['HOME'] = tmphome
        try:
            cfgdir = os.path.join(tmphome, ".config", "arvados")
            os.makedirs(cfgdir)
            with open(os.path.join(cfgdir, "zzzzz.conf"), "wt") as f:
                f.write("ARVADOS_API_HOST=%s\n" % os.environ["ARVADOS_API_HOST"])
                f.write("ARVADOS_API_TOKEN=%s\n" % os.environ["ARVADOS_API_TOKEN"])
                f.write("ARVADOS_API_HOST_INSECURE=1\n")

            contents = api.groups().list(filters=[["owner_uuid", "=", dest_proj]]).execute()
            assert len(contents["items"]) == 0

            with tutil.redirected_streams(
                    stdout=tutil.StringIO, stderr=tutil.StringIO) as (out, err):
                try:
                    self.run_copy(["--project-uuid", dest_proj, "--storage-classes", "foo", src_proj])
                except SystemExit as e:
                    assert e.code == 0
                copy_uuid_from_stdout = out.getvalue().strip()

            contents = api.groups().list(filters=[["owner_uuid", "=", dest_proj]]).execute()
            assert len(contents["items"]) == 1

            assert contents["items"][0]["name"] == "arv-copy project"
            copied_project = contents["items"][0]["uuid"]

            assert copied_project == copy_uuid_from_stdout

            contents = api.collections().list(filters=[["owner_uuid", "=", copied_project]]).execute()
            assert len(contents["items"]) == 1

            assert contents["items"][0]["uuid"] != c.manifest_locator()
            assert contents["items"][0]["name"] == "arv-copy foo collection"
            assert contents["items"][0]["portable_data_hash"] == c.portable_data_hash()
            assert contents["items"][0]["storage_classes_desired"] == ["foo"]

        finally:
            os.environ['HOME'] = home_was
            shutil.rmtree(tmphome)



#!/usr/bin/env python3
# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

#
# Migration tool for merging user accounts belonging to the same user
# but on separate clusters to use a single user account managed by a
# specific cluster.
#
# If you're working on this, see
# arvados/sdk/python/tests/fed-migrate/README for information about
# the testing infrastructure.

import arvados
import arvados.util
import arvados.errors
import csv
import sys
import argparse
import hmac
import urllib.parse
import os
import hashlib
import re
from arvados._version import __version__

EMAIL=0
USERNAME=1
UUID=2
HOMECLUSTER=3

def connect_clusters(args):
    clusters = {}
    errors = []
    loginCluster = None
    if args.tokens:
        print("Reading %s" % args.tokens)
        with open(args.tokens, "rt") as f:
            for r in csv.reader(f):
                if len(r) != 2:
                    continue
                host = r[0]
                token = r[1]
                print("Contacting %s" % (host))
                arv = arvados.api(host=host, token=token, cache=False)
                clusters[arv._rootDesc["uuidPrefix"]] = arv
    else:
        arv = arvados.api(cache=False)
        rh = arv._rootDesc["remoteHosts"]
        tok = arv.api_client_authorizations().current().execute()
        token = "v2/%s/%s" % (tok["uuid"], tok["api_token"])

        for k,v in rh.items():
            arv = arvados.api(host=v, token=token, cache=False, insecure=os.environ.get("ARVADOS_API_HOST_INSECURE"))
            clusters[k] = arv

    for _, arv in clusters.items():
        config = arv.configs().get().execute()
        if config["Login"]["LoginCluster"] != "" and loginCluster is None:
            loginCluster = config["Login"]["LoginCluster"]

    print("Checking that the federation is well connected")
    for arv in clusters.values():
        config = arv.configs().get().execute()
        if loginCluster and config["Login"]["LoginCluster"] != loginCluster and config["ClusterID"] != loginCluster:
            errors.append("Inconsistent login cluster configuration, expected '%s' on %s but was '%s'" % (loginCluster, config["ClusterID"], config["Login"]["LoginCluster"]))
            continue

        if arv._rootDesc["revision"] < "20200331":
            errors.append("Arvados API server revision on cluster '%s' is too old, must be updated to at least Arvados 2.0.2 before running migration." % config["ClusterID"])
            continue

        try:
            cur = arv.users().current().execute()
        except arvados.errors.ApiError as e:
            errors.append("checking token for %s   %s" % (arv._rootDesc["rootUrl"], e))
            continue

        if not cur["is_admin"]:
            errors.append("User %s is not admin on %s" % (cur["uuid"], arv._rootDesc["uuidPrefix"]))
            continue

        for r in clusters:
            if r != arv._rootDesc["uuidPrefix"] and r not in arv._rootDesc["remoteHosts"]:
                errors.append("%s is missing from remoteHosts of %s" % (r, arv._rootDesc["uuidPrefix"]))
        for r in arv._rootDesc["remoteHosts"]:
            if r != "*" and r not in clusters:
                print("WARNING: %s is federated with %s but %s is missing from the tokens file or the token is invalid" % (arv._rootDesc["uuidPrefix"], r, r))

    return clusters, errors, loginCluster


def fetch_users(clusters, loginCluster):
    rows = []
    by_email = {}
    by_username = {}

    users = []
    for c, arv in clusters.items():
        print("Getting user list from %s" % c)
        ul = arvados.util.list_all(arv.users().list, bypass_federation=True)
        for l in ul:
            if l["uuid"].startswith(c):
                users.append(l)

    # Users list is sorted by email
    # Go through users and collect users with same email
    # when we see a different email (or get to the end)
    # call add_accum_rows() to generate the report rows with
    # the "home cluster" set, and also fill in the by_email table.

    users = sorted(users, key=lambda u: u["email"]+"::"+(u["username"] or "")+"::"+u["uuid"])

    accum = []
    lastemail = None

    def add_accum_rows():
        homeuuid = None
        for a in accum:
            uuids = set(a["uuid"] for a in accum)
            homeuuid = ((len(uuids) == 1) and uuids.pop()) or ""
        for a in accum:
            r = (a["email"], a["username"], a["uuid"], loginCluster or homeuuid[0:5])
            by_email.setdefault(a["email"], {})
            by_email[a["email"]][a["uuid"]] = r
            homeuuid_and_username = "%s::%s" % (r[HOMECLUSTER], a["username"])
            if homeuuid_and_username not in by_username:
                by_username[homeuuid_and_username] = a["email"]
            elif by_username[homeuuid_and_username] != a["email"]:
                print("ERROR: the username '%s' is listed for both '%s' and '%s' on cluster '%s'" % (r[USERNAME], r[EMAIL], by_username[homeuuid_and_username], r[HOMECLUSTER]))
                exit(1)
            rows.append(r)

    for u in users:
        if u["uuid"].endswith("-anonymouspublic") or u["uuid"].endswith("-000000000000000"):
            continue
        if lastemail == None:
            lastemail = u["email"]
        if u["email"] == lastemail:
            accum.append(u)
        else:
            add_accum_rows()
            lastemail = u["email"]
            accum = [u]

    add_accum_rows()

    return rows, by_email, by_username


def read_migrations(args, by_email, by_username):
    rows = []
    with open(args.migrate or args.dry_run, "rt") as f:
        for r in csv.reader(f):
            if r[EMAIL] == "email":
                continue
            by_email.setdefault(r[EMAIL], {})
            by_email[r[EMAIL]][r[UUID]] = r

            homeuuid_and_username = "%s::%s" % (r[HOMECLUSTER], r[USERNAME])
            if homeuuid_and_username not in by_username:
                by_username[homeuuid_and_username] = r[EMAIL]
            elif by_username[homeuuid_and_username] != r[EMAIL]:
                print("ERROR: the username '%s' is listed for both '%s' and '%s' on cluster '%s'" % (r[USERNAME], r[EMAIL], by_username[homeuuid_and_username], r[HOMECLUSTER]))
                exit(1)

            rows.append(r)
    return rows

def update_username(args, email, user_uuid, username, migratecluster, migratearv):
    print("(%s) Updating username of %s to '%s' on %s" % (email, user_uuid, username, migratecluster))
    if args.dry_run:
        return
    try:
        conflicts = migratearv.users().list(filters=[["username", "=", username]], bypass_federation=True).execute()
        if conflicts["items"]:
            # There's already a user with the username, move the old user out of the way
            migratearv.users().update(uuid=conflicts["items"][0]["uuid"],
                                        bypass_federation=True,
                                        body={"user": {"username": username+"migrate"}}).execute()
        migratearv.users().update(uuid=user_uuid,
                                    bypass_federation=True,
                                    body={"user": {"username": username}}).execute()
    except arvados.errors.ApiError as e:
        print("(%s) Error updating username of %s to '%s' on %s: %s" % (email, user_uuid, username, migratecluster, e))


def choose_new_user(args, by_email, email, userhome, username, old_user_uuid, clusters):
    candidates = []
    conflict = False
    for b in by_email[email].values():
        if b[2].startswith(userhome):
            candidates.append(b)
        if b[1] != username and b[3] == userhome:
            print("(%s) Cannot migrate %s, conflicting usernames %s and %s" % (email, old_user_uuid, b[1], username))
            conflict = True
            break
    if conflict:
        return None
    if len(candidates) == 0:
        if len(userhome) == 5 and userhome not in clusters:
            print("(%s) Cannot migrate %s, unknown home cluster %s (typo?)" % (email, old_user_uuid, userhome))
            return None
        print("(%s) No user listed with same email to migrate %s to %s, will create new user with username '%s'" % (email, old_user_uuid, userhome, username))
        if not args.dry_run:
            oldhomecluster = old_user_uuid[0:5]
            oldhomearv = clusters[oldhomecluster]
            newhomecluster = userhome[0:5]
            homearv = clusters[userhome]
            user = None
            try:
                olduser = oldhomearv.users().get(uuid=old_user_uuid).execute()
                conflicts = homearv.users().list(filters=[["username", "=", username]],
                                                 bypass_federation=True).execute()
                if conflicts["items"]:
                    homearv.users().update(
                        uuid=conflicts["items"][0]["uuid"],
                        bypass_federation=True,
                        body={"user": {"username": username+"migrate"}}).execute()
                user = homearv.users().create(
                    body={"user": {
                        "email": email,
                        "first_name": olduser["first_name"],
                        "last_name": olduser["last_name"],
                        "username": username,
                        "is_active": olduser["is_active"]}}).execute()
            except arvados.errors.ApiError as e:
                print("(%s) Could not create user: %s" % (email, str(e)))
                return None

            tup = (email, username, user["uuid"], userhome)
        else:
            # dry run
            tup = (email, username, "%s-tpzed-xfakexfakexfake" % (userhome[0:5]), userhome)
        by_email[email][tup[2]] = tup
        candidates.append(tup)
    if len(candidates) > 1:
        print("(%s) Multiple users listed to migrate %s to %s, use full uuid" % (email, old_user_uuid, userhome))
        return None
    return candidates[0][2]


def activate_remote_user(args, email, homearv, migratearv, old_user_uuid, new_user_uuid):
    # create a token for the new user and salt it for the
    # migration cluster, then use it to access the migration
    # cluster as the new user once before merging to ensure
    # the new user is known on that cluster.
    migratecluster = migratearv._rootDesc["uuidPrefix"]
    try:
        if not args.dry_run:
            newtok = homearv.api_client_authorizations().create(body={
                "api_client_authorization": {'owner_uuid': new_user_uuid}}).execute()
        else:
            newtok = {"uuid": "dry-run", "api_token": "12345"}
    except arvados.errors.ApiError as e:
        print("(%s) Could not create API token for %s: %s" % (email, new_user_uuid, e))
        return None

    try:
        findolduser = migratearv.users().list(filters=[["uuid", "=", old_user_uuid]], bypass_federation=True).execute()
        if len(findolduser["items"]) == 0:
            return False
        if len(findolduser["items"]) == 1:
            olduser = findolduser["items"][0]
        else:
            print("(%s) Unexpected result" % (email))
            return None
    except arvados.errors.ApiError as e:
        print("(%s) Could not retrieve user %s from %s, user may have already been migrated: %s" % (email, old_user_uuid, migratecluster, e))
        return None

    salted = 'v2/' + newtok["uuid"] + '/' + hmac.new(newtok["api_token"].encode(),
                                                     msg=migratecluster.encode(),
                                                     digestmod=hashlib.sha1).hexdigest()
    try:
        ru = urllib.parse.urlparse(migratearv._rootDesc["rootUrl"])
        if not args.dry_run:
            newuser = arvados.api(host=ru.netloc, token=salted,
                                  insecure=os.environ.get("ARVADOS_API_HOST_INSECURE")).users().current().execute()
        else:
            newuser = {"is_active": True, "username": email.split('@')[0], "is_admin": False}
    except arvados.errors.ApiError as e:
        print("(%s) Error getting user info for %s from %s: %s" % (email, new_user_uuid, migratecluster, e))
        return None

    if not newuser["is_active"] and olduser["is_active"]:
        print("(%s) Activating user %s on %s" % (email, new_user_uuid, migratecluster))
        try:
            if not args.dry_run:
                migratearv.users().update(uuid=new_user_uuid, bypass_federation=True,
                                          body={"is_active": True}).execute()
        except arvados.errors.ApiError as e:
            print("(%s) Could not activate user %s on %s: %s" % (email, new_user_uuid, migratecluster, e))
            return None

    if olduser["is_admin"] and not newuser["is_admin"]:
        print("(%s) Not migrating %s because user is admin but target user %s is not admin on %s. Please ensure the user admin status is the same on both clusters. Note that a federated admin account has admin privileges on the entire federation." % (email, old_user_uuid, new_user_uuid, migratecluster))
        return None

    return newuser

def migrate_user(args, migratearv, email, new_user_uuid, old_user_uuid):
    if args.dry_run:
        return
    try:
        new_owner_uuid = new_user_uuid
        if args.data_into_subproject:
            grp = migratearv.groups().create(body={
                "owner_uuid": new_user_uuid,
                "name": "Migrated from %s (%s)" % (email, old_user_uuid),
                "group_class": "project"
            }, ensure_unique_name=True).execute()
            new_owner_uuid = grp["uuid"]
        migratearv.users().merge(old_user_uuid=old_user_uuid,
                                    new_user_uuid=new_user_uuid,
                                    new_owner_uuid=new_owner_uuid,
                                    redirect_to_new_user=True).execute()
    except arvados.errors.ApiError as e:
        name_collision = re.search(r'Key \(owner_uuid, name\)=\((.*?), (.*?)\) already exists\.\n.*UPDATE "(.*?)"', e._get_reason())
        if name_collision:
            target_owner, rsc_name, rsc_type = name_collision.groups()
            print("(%s) Cannot migrate to %s because both origin and target users have a %s named '%s'. Please rename the conflicting items or use --data-into-subproject to migrate all users' data into a special subproject." % (email, target_owner, rsc_type[:-1], rsc_name))
        else:
            print("(%s) Skipping user migration because of error: %s" % (email, e))


def main():
    parser = argparse.ArgumentParser(description='Migrate users to federated identity, see https://doc.arvados.org/admin/merge-remote-account.html')
    parser.add_argument(
        '--version', action='version', version="%s %s" % (sys.argv[0], __version__),
        help='Print version and exit.')
    parser.add_argument('--tokens', type=str, metavar='FILE', required=False, help="Read tokens from FILE. Not needed when using LoginCluster.")
    parser.add_argument('--data-into-subproject', action="store_true", help="Migrate user's data into a separate subproject. This can be used to avoid name collisions from within an account.")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--report', type=str, metavar='FILE', help="Generate report .csv file listing users by email address and their associated Arvados accounts.")
    group.add_argument('--migrate', type=str, metavar='FILE', help="Consume report .csv and migrate users to designated Arvados accounts.")
    group.add_argument('--dry-run', type=str, metavar='FILE', help="Consume report .csv and report how user would be migrated to designated Arvados accounts.")
    group.add_argument('--check', action="store_true", help="Check that tokens are usable and the federation is well connected.")
    args = parser.parse_args()

    clusters, errors, loginCluster = connect_clusters(args)

    if errors:
        for e in errors:
            print("ERROR: "+str(e))
        exit(1)

    if args.check:
        print("Tokens file passed checks")
        exit(0)

    rows, by_email, by_username = fetch_users(clusters, loginCluster)

    if args.report:
        out = csv.writer(open(args.report, "wt"))
        out.writerow(("email", "username", "user uuid", "home cluster"))
        for r in rows:
            out.writerow(r)
        print("Wrote %s" % args.report)
        return

    if args.migrate or args.dry_run:
        if args.dry_run:
            print("Performing dry run")

        rows = read_migrations(args, by_email, by_username)

        for r in rows:
            email = r[EMAIL]
            username = r[USERNAME]
            old_user_uuid = r[UUID]
            userhome = r[HOMECLUSTER]

            if userhome == "":
                print("(%s) Skipping %s, no home cluster specified" % (email, old_user_uuid))
            if old_user_uuid.startswith(userhome):
                migratecluster = old_user_uuid[0:5]
                migratearv = clusters[migratecluster]
                if migratearv.users().get(uuid=old_user_uuid).execute()["username"] != username:
                    update_username(args, email, old_user_uuid, username, migratecluster, migratearv)
                continue

            new_user_uuid = choose_new_user(args, by_email, email, userhome, username, old_user_uuid, clusters)
            if new_user_uuid is None:
                continue

            remote_users = {}
            got_error = False
            for migratecluster in clusters:
                # cluster where the migration is happening
                migratearv = clusters[migratecluster]

                # the user's new home cluster
                newhomecluster = userhome[0:5]
                homearv = clusters[newhomecluster]

                newuser = activate_remote_user(args, email, homearv, migratearv, old_user_uuid, new_user_uuid)
                if newuser is None:
                    got_error = True
                remote_users[migratecluster] = newuser

            if not got_error:
                for migratecluster in clusters:
                    migratearv = clusters[migratecluster]
                    newuser = remote_users[migratecluster]
                    if newuser is False:
                        continue

                    print("(%s) Migrating %s to %s on %s" % (email, old_user_uuid, new_user_uuid, migratecluster))

                    migrate_user(args, migratearv, email, new_user_uuid, old_user_uuid)

                    if newuser['username'] != username:
                        update_username(args, email, new_user_uuid, username, migratecluster, migratearv)

if __name__ == "__main__":
    main()



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
from . import config

import re

def escape(path):
    path = re.sub('\\\\', lambda m: '\\134', path)
    path = re.sub('[:\000-\040]', lambda m: "\\%03o" % ord(m.group(0)), path)
    return path

def normalize_stream(stream_name, stream):
    """Take manifest stream and return a list of tokens in normalized format.

    :stream_name:
      The name of the stream.

    :stream:
      A dict mapping each filename to a list of `_range.LocatorAndRange` objects.

    """

    stream_name = escape(stream_name)
    stream_tokens = [stream_name]
    sortedfiles = list(stream.keys())
    sortedfiles.sort()

    blocks = {}
    streamoffset = 0
    # Go through each file and add each referenced block exactly once.
    for streamfile in sortedfiles:
        for segment in stream[streamfile]:
            if segment.locator not in blocks:
                stream_tokens.append(segment.locator)
                blocks[segment.locator] = streamoffset
                streamoffset += segment.block_size

    # Add the empty block if the stream is otherwise empty.
    if len(stream_tokens) == 1:
        stream_tokens.append(config.EMPTY_BLOCK_LOCATOR)

    for streamfile in sortedfiles:
        # Add in file segments
        current_span = None
        fout = escape(streamfile)
        for segment in stream[streamfile]:
            # Collapse adjacent segments
            streamoffset = blocks[segment.locator] + segment.segment_offset
            if current_span is None:
                current_span = [streamoffset, streamoffset + segment.segment_size]
            else:
                if streamoffset == current_span[1]:
                    current_span[1] += segment.segment_size
                else:
                    stream_tokens.append(u"{0}:{1}:{2}".format(current_span[0], current_span[1] - current_span[0], fout))
                    current_span = [streamoffset, streamoffset + segment.segment_size]

        if current_span is not None:
            stream_tokens.append(u"{0}:{1}:{2}".format(current_span[0], current_span[1] - current_span[0], fout))

        if not stream[streamfile]:
            stream_tokens.append(u"0:0:{0}".format(fout))

    return stream_tokens



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
from builtins import hex
from builtins import str
from builtins import range
from builtins import object
import datetime
import mock
import os
import unittest
import time

import arvados
from arvados._ranges import Range
from arvados.keep import KeepLocator
from arvados.collection import Collection
from arvados.arvfile import ArvadosFile, ArvadosFileReader

from . import arvados_testutil as tutil
from .test_stream import StreamFileReaderTestCase, StreamRetryTestMixin

class ArvadosFileWriterTestCase(unittest.TestCase):
    class MockKeep(object):
        def __init__(self, blocks):
            self.blocks = blocks
            self.requests = []
        def get(self, locator, num_retries=0, prefetch=False):
            self.requests.append(locator)
            return self.blocks.get(locator)
        def get_from_cache(self, locator):
            self.requests.append(locator)
            return self.blocks.get(locator)
        def put(self, data, num_retries=None, copies=None, classes=[]):
            pdh = tutil.str_keep_locator(data)
            self.blocks[pdh] = bytes(data)
            return pdh

    class MockApi(object):
        def __init__(self, b, r):
            self.body = b
            self.response = r
            self._schema = ArvadosFileWriterTestCase.MockApi.MockSchema()
            self._rootDesc = {}
        class MockSchema(object):
            def __init__(self):
                self.schemas = {'Collection': {'properties': {'replication_desired': {'type':'integer'}}}}
        class MockCollections(object):
            def __init__(self, b, r):
                self.body = b
                self.response = r
            class Execute(object):
                def __init__(self, r):
                    self.response = r
                def execute(self, num_retries=None):
                    return self.response
            def create(self, ensure_unique_name=False, body=None):
                if body != self.body:
                    raise Exception("Body %s does not match expectation %s" % (body, self.body))
                return ArvadosFileWriterTestCase.MockApi.MockCollections.Execute(self.response)
            def update(self, uuid=None, body=None):
                return ArvadosFileWriterTestCase.MockApi.MockCollections.Execute(self.response)
        def collections(self):
            return ArvadosFileWriterTestCase.MockApi.MockCollections(self.body, self.response)


    def test_truncate(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "781e5e245d69b566979b86e28d23f2c7+10": b"0123456789",
        })
        api = ArvadosFileWriterTestCase.MockApi({
            "name": "test_truncate",
            "manifest_text": ". 781e5e245d69b566979b86e28d23f2c7+10 0:8:count.txt\n",
            "replication_desired": None,
        }, {
            "uuid": "zzzzz-4zz18-mockcollection0",
            "manifest_text": ". 781e5e245d69b566979b86e28d23f2c7+10 0:8:count.txt\n",
            "portable_data_hash":"7fcd0eaac3aad4c31a6a0e756475da92+52",
        })
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n',
                        api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(writer.size(), 10)
            self.assertEqual(b"0123456789", writer.read(12))

            writer.truncate(8)

            # Make sure reading off the end doesn't break
            self.assertEqual(b"", writer.read(12))

            self.assertEqual(writer.size(), 8)
            writer.seek(0, os.SEEK_SET)
            self.assertEqual(b"01234567", writer.read(12))

            self.assertIsNone(c.manifest_locator())
            self.assertTrue(c.modified())
            c.save_new("test_truncate")
            self.assertEqual("zzzzz-4zz18-mockcollection0", c.manifest_locator())
            self.assertFalse(c.modified())


    def test_truncate2(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "781e5e245d69b566979b86e28d23f2c7+10": b"0123456789",
        })
        api = ArvadosFileWriterTestCase.MockApi({
            "name": "test_truncate2",
            "manifest_text": ". 781e5e245d69b566979b86e28d23f2c7+10 7f614da9329cd3aebf59b91aadc30bf0+67108864 0:12:count.txt\n",
            "replication_desired": None,
        }, {
            "uuid": "zzzzz-4zz18-mockcollection0",
            "manifest_text": ". 781e5e245d69b566979b86e28d23f2c7+10 7f614da9329cd3aebf59b91aadc30bf0+67108864 0:12:count.txt\n",
            "portable_data_hash": "272da898abdf86ddc71994835e3155f8+95",
        })
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n',
                        api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(writer.size(), 10)
            self.assertEqual(b"0123456789", writer.read(12))

            # extend file size
            writer.truncate(12)

            self.assertEqual(writer.size(), 12)
            writer.seek(0, os.SEEK_SET)
            self.assertEqual(b"0123456789\x00\x00", writer.read(12))

            self.assertIsNone(c.manifest_locator())
            self.assertTrue(c.modified())
            c.save_new("test_truncate2")
            self.assertEqual("zzzzz-4zz18-mockcollection0", c.manifest_locator())
            self.assertFalse(c.modified())

    def test_truncate3(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "781e5e245d69b566979b86e28d23f2c7+10": b"0123456789",
            "a925576942e94b2ef57a066101b48876+10": b"abcdefghij",
        })
        api = ArvadosFileWriterTestCase.MockApi({
            "name": "test_truncate",
            "manifest_text": ". 781e5e245d69b566979b86e28d23f2c7+10 0:8:count.txt\n",
            "replication_desired": None,
        }, {
            "uuid": "zzzzz-4zz18-mockcollection0",
            "manifest_text": ". 781e5e245d69b566979b86e28d23f2c7+10 0:8:count.txt\n",
            "portable_data_hash": "7fcd0eaac3aad4c31a6a0e756475da92+52",
        })
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 a925576942e94b2ef57a066101b48876+10 0:20:count.txt\n',
                        api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(writer.size(), 20)
            self.assertEqual(b"0123456789ab", writer.read(12))
            self.assertEqual(12, writer.tell())

            writer.truncate(8)

            # Make sure reading off the end doesn't break
            self.assertEqual(12, writer.tell())
            self.assertEqual(b"", writer.read(12))

            self.assertEqual(writer.size(), 8)
            self.assertEqual(2, writer.seek(-10, os.SEEK_CUR))
            self.assertEqual(b"234567", writer.read(12))

            self.assertIsNone(c.manifest_locator())
            self.assertTrue(c.modified())
            c.save_new("test_truncate")
            self.assertEqual("zzzzz-4zz18-mockcollection0", c.manifest_locator())
            self.assertFalse(c.modified())

    def test_write_to_end(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "781e5e245d69b566979b86e28d23f2c7+10": b"0123456789",
        })
        api = ArvadosFileWriterTestCase.MockApi({
            "name": "test_append",
            "manifest_text": ". 781e5e245d69b566979b86e28d23f2c7+10 acbd18db4cc2f85cedef654fccc4a4d8+3 0:13:count.txt\n",
            "replication_desired": None,
        }, {
            "uuid": "zzzzz-4zz18-mockcollection0",
            "manifest_text": ". 781e5e245d69b566979b86e28d23f2c7+10 acbd18db4cc2f85cedef654fccc4a4d8+3 0:13:count.txt\n",
            "portable_data_hash": "c5c3af76565c8efb6a806546bcf073f3+88",
        })
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n',
                             api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(writer.size(), 10)

            self.assertEqual(5, writer.seek(5, os.SEEK_SET))
            self.assertEqual(b"56789", writer.read(8))

            writer.seek(10, os.SEEK_SET)
            writer.write("foo")
            self.assertEqual(writer.size(), 13)

            writer.seek(5, os.SEEK_SET)
            self.assertEqual(b"56789foo", writer.read(8))

            self.assertIsNone(c.manifest_locator())
            self.assertTrue(c.modified())
            self.assertIsNone(keep.get("acbd18db4cc2f85cedef654fccc4a4d8+3"))

            c.save_new("test_append")
            self.assertEqual("zzzzz-4zz18-mockcollection0", c.manifest_locator())
            self.assertFalse(c.modified())
            self.assertEqual(b"foo", keep.get("acbd18db4cc2f85cedef654fccc4a4d8+3"))


    def test_append(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "781e5e245d69b566979b86e28d23f2c7+10": b"0123456789",
        })
        for (mode, convert) in (
                ('a+', lambda data: data.decode(encoding='utf-8')),
                ('at+', lambda data: data.decode(encoding='utf-8')),
                ('ab+', lambda data: data)):
            c = Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n', keep_client=keep)
            writer = c.open("count.txt", mode)
            self.assertEqual(writer.read(20), convert(b"0123456789"))

            writer.seek(0, os.SEEK_SET)
            writer.write(convert(b"hello"))
            self.assertEqual(writer.read(), convert(b""))
            if 'b' in mode:
                writer.seek(-5, os.SEEK_CUR)
                self.assertEqual(writer.read(3), convert(b"hel"))
                self.assertEqual(writer.read(), convert(b"lo"))
            else:
                with self.assertRaises(IOError):
                    writer.seek(-5, os.SEEK_CUR)
                with self.assertRaises(IOError):
                    writer.seek(-3, os.SEEK_END)
            writer.seek(0, os.SEEK_SET)
            writer.read(7)
            self.assertEqual(7, writer.tell())
            self.assertEqual(7, writer.seek(7, os.SEEK_SET))

            writer.seek(0, os.SEEK_SET)
            self.assertEqual(writer.read(), convert(b"0123456789hello"))

            writer.seek(0)
            writer.write(convert(b"world"))
            self.assertEqual(writer.read(), convert(b""))
            writer.seek(0)
            self.assertEqual(writer.read(), convert(b"0123456789helloworld"))

            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 fc5e038d38a57032085441e7fe7010b0+10 0:20:count.txt\n", c.portable_manifest_text())

    def test_write_at_beginning(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "781e5e245d69b566979b86e28d23f2c7+10": b"0123456789",
        })
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n',
                             keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(b"0123456789", writer.readfrom(0, 13))
            writer.seek(0, os.SEEK_SET)
            writer.write("foo")
            self.assertEqual(writer.size(), 10)
            self.assertEqual(b"foo3456789", writer.readfrom(0, 13))
            self.assertEqual(". acbd18db4cc2f85cedef654fccc4a4d8+3 781e5e245d69b566979b86e28d23f2c7+10 0:3:count.txt 6:7:count.txt\n", c.portable_manifest_text())

    def test_write_empty(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        with Collection(keep_client=keep) as c:
            writer = c.open("count.txt", "wb")
            self.assertEqual(writer.size(), 0)
            self.assertEqual(". d41d8cd98f00b204e9800998ecf8427e+0 0:0:count.txt\n", c.portable_manifest_text())

    def test_save_manifest_text(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        with Collection(keep_client=keep) as c:
            writer = c.open("count.txt", "wb")
            writer.write(b"0123456789")
            self.assertEqual('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n', c.portable_manifest_text())
            self.assertNotIn('781e5e245d69b566979b86e28d23f2c7+10', keep.blocks)

            self.assertEqual('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n', c.save_new(create_collection_record=False))
            self.assertIn('781e5e245d69b566979b86e28d23f2c7+10', keep.blocks)

    def test_get_manifest_text_commits(self):
         keep = ArvadosFileWriterTestCase.MockKeep({})
         with Collection(keep_client=keep) as c:
             writer = c.open("count.txt", "wb")
             writer.write("0123456789")
             self.assertEqual('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n', c.portable_manifest_text())
             self.assertNotIn('781e5e245d69b566979b86e28d23f2c7+10', keep.blocks)
             self.assertEqual('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n', c.manifest_text())
             self.assertIn('781e5e245d69b566979b86e28d23f2c7+10', keep.blocks)


    def test_write_in_middle(self):
        keep = ArvadosFileWriterTestCase.MockKeep({"781e5e245d69b566979b86e28d23f2c7+10": b"0123456789"})
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n',
                             keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(b"0123456789", writer.readfrom(0, 13))
            writer.seek(3, os.SEEK_SET)
            writer.write("foo")
            self.assertEqual(writer.size(), 10)
            self.assertEqual(b"012foo6789", writer.readfrom(0, 13))
            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 acbd18db4cc2f85cedef654fccc4a4d8+3 0:3:count.txt 10:3:count.txt 6:4:count.txt\n", c.portable_manifest_text())

    def test_write_at_end(self):
        keep = ArvadosFileWriterTestCase.MockKeep({"781e5e245d69b566979b86e28d23f2c7+10": b"0123456789"})
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n',
                             keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(b"0123456789", writer.readfrom(0, 13))
            writer.seek(7, os.SEEK_SET)
            writer.write("foo")
            self.assertEqual(writer.size(), 10)
            self.assertEqual(b"0123456foo", writer.readfrom(0, 13))
            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 acbd18db4cc2f85cedef654fccc4a4d8+3 0:7:count.txt 10:3:count.txt\n", c.portable_manifest_text())

    def test_write_across_segment_boundary(self):
        keep = ArvadosFileWriterTestCase.MockKeep({"781e5e245d69b566979b86e28d23f2c7+10": b"0123456789"})
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt 0:10:count.txt\n',
                             keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(b"012345678901234", writer.readfrom(0, 15))
            writer.seek(7, os.SEEK_SET)
            writer.write("foobar")
            self.assertEqual(writer.size(), 20)
            self.assertEqual(b"0123456foobar34", writer.readfrom(0, 15))
            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 3858f62230ac3c915f300c664312c63f+6 0:7:count.txt 10:6:count.txt 3:7:count.txt\n", c.portable_manifest_text())

    def test_write_across_several_segments(self):
        keep = ArvadosFileWriterTestCase.MockKeep({"781e5e245d69b566979b86e28d23f2c7+10": b"0123456789"})
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:4:count.txt 0:4:count.txt 0:4:count.txt',
                             keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(b"012301230123", writer.readfrom(0, 15))
            writer.seek(2, os.SEEK_SET)
            writer.write("abcdefg")
            self.assertEqual(writer.size(), 12)
            self.assertEqual(b"01abcdefg123", writer.readfrom(0, 15))
            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 7ac66c0f148de9519b8bd264312c4d64+7 0:2:count.txt 10:7:count.txt 1:3:count.txt\n", c.portable_manifest_text())

    def test_write_large(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({"name":"test_write_large",
                                                 "manifest_text": ". a5de24f4417cfba9d5825eadc2f4ca49+67108000 598cc1a4ccaef8ab6e4724d87e675d78+32892000 0:100000000:count.txt\n",
                                                 "replication_desired":None},
                                                {"uuid":"zzzzz-4zz18-mockcollection0",
                                                 "manifest_text": ". a5de24f4417cfba9d5825eadc2f4ca49+67108000 598cc1a4ccaef8ab6e4724d87e675d78+32892000 0:100000000:count.txt\n",
                                                 "portable_data_hash":"9132ca8e3f671c76103a38f5bc24328c+108"})
        with Collection('. ' + arvados.config.EMPTY_BLOCK_LOCATOR + ' 0:0:count.txt',
                             api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            text = "0123456789" * 100
            for b in range(0, 100000):
                writer.write(text)
            self.assertEqual(writer.size(), 100000000)

            self.assertIsNone(c.manifest_locator())
            self.assertTrue(c.modified())
            c.save_new("test_write_large")
            self.assertEqual("zzzzz-4zz18-mockcollection0", c.manifest_locator())
            self.assertFalse(c.modified())

    def test_large_write(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({}, {})
        with Collection('. ' + arvados.config.EMPTY_BLOCK_LOCATOR + ' 0:0:count.txt',
                             api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(writer.size(), 0)

            text = "0123456789"
            writer.write(text)
            text = "0123456789" * 9999999
            writer.write(text)
            self.assertEqual(writer.size(), 100000000)

            self.assertEqual(c.manifest_text(), ". 781e5e245d69b566979b86e28d23f2c7+10 48dd23ea1645fd47d789804d71b5bb8e+67108864 77c57dc6ac5a10bb2205caaa73187994+32891126 0:100000000:count.txt\n")

    def test_sparse_write(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({}, {})
        with Collection('. ' + arvados.config.EMPTY_BLOCK_LOCATOR + ' 0:0:count.txt',
                             api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(writer.size(), 0)

            text = b"0123456789"
            writer.seek(2)
            writer.write(text)
            self.assertEqual(writer.size(), 12)
            writer.seek(0, os.SEEK_SET)
            self.assertEqual(writer.read(), b"\x00\x00"+text)

            self.assertEqual(c.manifest_text(), ". 7f614da9329cd3aebf59b91aadc30bf0+67108864 781e5e245d69b566979b86e28d23f2c7+10 0:2:count.txt 67108864:10:count.txt\n")

    def test_sparse_write2(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({}, {})
        with Collection('. ' + arvados.config.EMPTY_BLOCK_LOCATOR + ' 0:0:count.txt',
                             api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            self.assertEqual(writer.size(), 0)

            text = "0123456789"
            writer.seek((arvados.config.KEEP_BLOCK_SIZE*2) + 2)
            writer.write(text)
            self.assertEqual(writer.size(), (arvados.config.KEEP_BLOCK_SIZE*2) + 12)
            writer.seek(0, os.SEEK_SET)

            self.assertEqual(c.manifest_text(), ". 7f614da9329cd3aebf59b91aadc30bf0+67108864 781e5e245d69b566979b86e28d23f2c7+10 0:67108864:count.txt 0:67108864:count.txt 0:2:count.txt 67108864:10:count.txt\n")

    def test_sparse_write3(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({}, {})
        for r in [[0, 1, 2, 3, 4], [4, 3, 2, 1, 0], [3, 2, 0, 4, 1]]:
            with Collection() as c:
                writer = c.open("count.txt", "rb+")
                self.assertEqual(writer.size(), 0)

                for i in r:
                    w = ("%s" % i) * 10
                    writer.seek(i*10)
                    writer.write(w.encode())
                writer.seek(0)
                self.assertEqual(writer.read(), b"00000000001111111111222222222233333333334444444444")

    def test_sparse_write4(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({}, {})
        for r in [[0, 1, 2, 4], [4, 2, 1, 0], [2, 0, 4, 1]]:
            with Collection() as c:
                writer = c.open("count.txt", "rb+")
                self.assertEqual(writer.size(), 0)

                for i in r:
                    w = ("%s" % i) * 10
                    writer.seek(i*10)
                    writer.write(w.encode())
                writer.seek(0)
                self.assertEqual(writer.read(), b"000000000011111111112222222222\x00\x00\x00\x00\x00\x00\x00\x00\x00\x004444444444")

    def test_rewrite_on_empty_file(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        with Collection('. ' + arvados.config.EMPTY_BLOCK_LOCATOR + ' 0:0:count.txt',
                             keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            for b in range(0, 10):
                writer.seek(0, os.SEEK_SET)
                writer.write("0123456789")

            self.assertEqual(writer.size(), 10)
            self.assertEqual(b"0123456789", writer.readfrom(0, 20))
            self.assertEqual(". 7a08b07e84641703e5f2c836aa59a170+100 90:10:count.txt\n", c.portable_manifest_text())
            writer.flush()
            self.assertEqual(writer.size(), 10)
            self.assertEqual(b"0123456789", writer.readfrom(0, 20))
            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n", c.portable_manifest_text())

    def test_rewrite_append_existing_file(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "781e5e245d69b566979b86e28d23f2c7+10": b"0123456789",
        })
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt',
                             keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            for b in range(0, 10):
                writer.seek(10, os.SEEK_SET)
                writer.write("abcdefghij")

            self.assertEqual(writer.size(), 20)
            self.assertEqual(b"0123456789abcdefghij", writer.readfrom(0, 20))
            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 ae5f43bab79cf0be33f025fa97ae7398+100 0:10:count.txt 100:10:count.txt\n", c.portable_manifest_text())

            writer.arvadosfile.flush()
            self.assertEqual(writer.size(), 20)
            self.assertEqual(b"0123456789abcdefghij", writer.readfrom(0, 20))
            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 a925576942e94b2ef57a066101b48876+10 0:20:count.txt\n", c.portable_manifest_text())

    def test_rewrite_over_existing_file(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "781e5e245d69b566979b86e28d23f2c7+10": b"0123456789",
        })
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt',
                             keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            for b in range(0, 10):
                writer.seek(5, os.SEEK_SET)
                writer.write("abcdefghij")

            self.assertEqual(writer.size(), 15)
            self.assertEqual(b"01234abcdefghij", writer.readfrom(0, 20))
            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 ae5f43bab79cf0be33f025fa97ae7398+100 0:5:count.txt 100:10:count.txt\n", c.portable_manifest_text())

            writer.arvadosfile.flush()

            self.assertEqual(writer.size(), 15)
            self.assertEqual(b"01234abcdefghij", writer.readfrom(0, 20))
            self.assertEqual(". 781e5e245d69b566979b86e28d23f2c7+10 a925576942e94b2ef57a066101b48876+10 0:5:count.txt 10:10:count.txt\n", c.portable_manifest_text())

    def test_write_large_rewrite(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({"name":"test_write_large",
                                                 "manifest_text": ". 3dc0d4bc21f48060bedcb2c91af4f906+32892003 a5de24f4417cfba9d5825eadc2f4ca49+67108000 0:3:count.txt 32892006:67107997:count.txt 3:32892000:count.txt\n",
                                                 "replication_desired":None},
                                                {"uuid":"zzzzz-4zz18-mockcollection0",
                                                 "manifest_text": ". 3dc0d4bc21f48060bedcb2c91af4f906+32892003 a5de24f4417cfba9d5825eadc2f4ca49+67108000 0:3:count.txt 32892006:67107997:count.txt 3:32892000:count.txt\n",
                                                 "portable_data_hash":"217665c6b713e1b78dfba7ebd42344db+156"})
        with Collection('. ' + arvados.config.EMPTY_BLOCK_LOCATOR + ' 0:0:count.txt',
                             api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "rb+")
            text = b''.join([b"0123456789" for a in range(0, 100)])
            for b in range(0, 100000):
                writer.write(text)
            writer.seek(0, os.SEEK_SET)
            writer.write("foo")
            self.assertEqual(writer.size(), 100000000)

            self.assertIsNone(c.manifest_locator())
            self.assertTrue(c.modified())
            c.save_new("test_write_large")
            self.assertEqual("zzzzz-4zz18-mockcollection0", c.manifest_locator())
            self.assertFalse(c.modified())

    def test_create(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({
            "name":"test_create",
            "manifest_text":". 2e9ec317e197819358fbc43afca7d837+8 0:8:count.txt\n",
            "replication_desired":None,
        }, {
            "uuid":"zzzzz-4zz18-mockcollection0",
            "manifest_text":". 2e9ec317e197819358fbc43afca7d837+8 0:8:count.txt\n",
            "portable_data_hash":"7a461a8c58601798f690f8b368ac4423+51",
        })
        with Collection(api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "wb+")
            self.assertEqual(writer.size(), 0)
            writer.write("01234567")
            self.assertEqual(writer.size(), 8)

            self.assertIsNone(c.manifest_locator())
            self.assertTrue(c.modified())
            self.assertIsNone(keep.get("2e9ec317e197819358fbc43afca7d837+8"))
            c.save_new("test_create")
            self.assertEqual("zzzzz-4zz18-mockcollection0", c.manifest_locator())
            self.assertFalse(c.modified())
            self.assertEqual(b"01234567", keep.get("2e9ec317e197819358fbc43afca7d837+8"))


    def test_create_subdir(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({"name":"test_create",
                                                 "manifest_text":"./foo/bar 2e9ec317e197819358fbc43afca7d837+8 0:8:count.txt\n",
                                                 "replication_desired":None},
                                                {"uuid":"zzzzz-4zz18-mockcollection0",
                                                 "manifest_text":"./foo/bar 2e9ec317e197819358fbc43afca7d837+8 0:8:count.txt\n",
                                                 "portable_data_hash":"1b02aaa62528d28a5be41651cbb9d7c7+59"})
        with Collection(api_client=api, keep_client=keep) as c:
            self.assertIsNone(c.api_response())
            writer = c.open("foo/bar/count.txt", "wb+")
            writer.write("01234567")
            self.assertFalse(c.committed())
            c.save_new("test_create")
            self.assertTrue(c.committed())
            self.assertEqual(c.api_response(), api.response)

    def test_overwrite(self):
        keep = ArvadosFileWriterTestCase.MockKeep({"781e5e245d69b566979b86e28d23f2c7+10": "0123456789"})
        api = ArvadosFileWriterTestCase.MockApi({"name":"test_overwrite",
                                                 "manifest_text":". 2e9ec317e197819358fbc43afca7d837+8 0:8:count.txt\n",
                                                 "replication_desired":None},
                                                {"uuid":"zzzzz-4zz18-mockcollection0",
                                                 "manifest_text":". 2e9ec317e197819358fbc43afca7d837+8 0:8:count.txt\n",
                                                 "portable_data_hash":"7a461a8c58601798f690f8b368ac4423+51"})
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n',
                             api_client=api, keep_client=keep) as c:
            writer = c.open("count.txt", "wb+")
            self.assertEqual(writer.size(), 0)
            writer.write("01234567")
            self.assertEqual(writer.size(), 8)

            self.assertIsNone(c.manifest_locator())
            self.assertTrue(c.modified())
            c.save_new("test_overwrite")
            self.assertEqual("zzzzz-4zz18-mockcollection0", c.manifest_locator())
            self.assertFalse(c.modified())

    def test_file_not_found(self):
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n') as c:
            with self.assertRaises(IOError):
                writer = c.open("nocount.txt", "rb")

    def test_cannot_open_directory(self):
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count.txt\n') as c:
            with self.assertRaises(IOError):
                writer = c.open(".", "rb")

    def test_create_multiple(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        api = ArvadosFileWriterTestCase.MockApi({"name":"test_create_multiple",
                                                 "manifest_text":". 2e9ec317e197819358fbc43afca7d837+8 e8dc4081b13434b45189a720b77b6818+8 0:8:count1.txt 8:8:count2.txt\n",
                                                 "replication_desired":None},
                                                {"uuid":"zzzzz-4zz18-mockcollection0",
                                                 "manifest_text":". 2e9ec317e197819358fbc43afca7d837+8 e8dc4081b13434b45189a720b77b6818+8 0:8:count1.txt 8:8:count2.txt\n",
                                                 "portable_data_hash":"71e7bb6c00d31fc2b4364199fd97be08+102"})
        with Collection(api_client=api, keep_client=keep) as c:
            w1 = c.open("count1.txt", "wb")
            w2 = c.open("count2.txt", "wb")
            w1.write("01234567")
            w2.write("abcdefgh")
            self.assertEqual(w1.size(), 8)
            self.assertEqual(w2.size(), 8)

            self.assertIsNone(c.manifest_locator())
            self.assertTrue(c.modified())
            self.assertIsNone(keep.get("2e9ec317e197819358fbc43afca7d837+8"))
            c.save_new("test_create_multiple")
            self.assertEqual("zzzzz-4zz18-mockcollection0", c.manifest_locator())
            self.assertFalse(c.modified())
            self.assertEqual(b"01234567", keep.get("2e9ec317e197819358fbc43afca7d837+8"))


class ArvadosFileReaderTestCase(StreamFileReaderTestCase):
    class MockParent(object):
        class MockBlockMgr(object):
            def __init__(self, blocks, nocache):
                self.blocks = blocks
                self.nocache = nocache
                self.num_get_threads = 1

            def block_prefetch(self, loc):
                pass

            def get_block_contents(self, loc, num_retries=0, cache_only=False):
                if self.nocache and cache_only:
                    return None
                return self.blocks[loc]

        def __init__(self, blocks, nocache):
            self.blocks = blocks
            self.nocache = nocache
            self.lock = arvados.arvfile.NoopLock()

        def root_collection(self):
            return self

        def _my_block_manager(self):
            return ArvadosFileReaderTestCase.MockParent.MockBlockMgr(self.blocks, self.nocache)


    def make_count_reader(self, nocache=False):
        stream = []
        n = 0
        blocks = {}
        for d in [b'01234', b'34567', b'67890']:
            loc = tutil.str_keep_locator(d)
            blocks[loc] = d
            stream.append(Range(loc, n, len(d)))
            n += len(d)
        af = ArvadosFile(ArvadosFileReaderTestCase.MockParent(blocks, nocache), "count.txt", stream=stream, segments=[Range(1, 0, 3), Range(6, 3, 3), Range(11, 6, 3)])
        return ArvadosFileReader(af, mode="rb")

    def test_read_block_crossing_behavior(self):
        # read() needs to return all the data requested if possible, even if it
        # crosses uncached blocks: https://arvados.org/issues/5856
        sfile = self.make_count_reader(nocache=True)
        self.assertEqual(b'12345678', sfile.read(8))

    def test_successive_reads(self):
        # Override StreamFileReaderTestCase.test_successive_reads
        sfile = self.make_count_reader(nocache=True)
        self.assertEqual(b'1234', sfile.read(4))
        self.assertEqual(b'5678', sfile.read(4))
        self.assertEqual(b'9', sfile.read(4))
        self.assertEqual(b'', sfile.read(4))

    def test_tell_after_block_read(self):
        # Override StreamFileReaderTestCase.test_tell_after_block_read
        sfile = self.make_count_reader(nocache=True)
        self.assertEqual(b'12345678', sfile.read(8))
        self.assertEqual(8, sfile.tell())

    def test_prefetch(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "2e9ec317e197819358fbc43afca7d837+8": b"01234567",
            "e8dc4081b13434b45189a720b77b6818+8": b"abcdefgh",
        })
        with Collection(". 2e9ec317e197819358fbc43afca7d837+8 e8dc4081b13434b45189a720b77b6818+8 0:16:count.txt\n", keep_client=keep) as c:
            r = c.open("count.txt", "rb")
            self.assertEqual(b"0123", r.read(4))
        self.assertIn("2e9ec317e197819358fbc43afca7d837+8", keep.requests)
        self.assertIn("e8dc4081b13434b45189a720b77b6818+8", keep.requests)

    def test__eq__from_manifest(self):
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count1.txt') as c1:
            with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count1.txt') as c2:
                self.assertTrue(c1["count1.txt"] == c2["count1.txt"])
                self.assertFalse(c1["count1.txt"] != c2["count1.txt"])

    def test__eq__from_writes(self):
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count1.txt') as c1:
            with Collection() as c2:
                f = c2.open("count1.txt", "wb")
                f.write("0123456789")

                self.assertTrue(c1["count1.txt"] == c2["count1.txt"])
                self.assertFalse(c1["count1.txt"] != c2["count1.txt"])

    def test__ne__(self):
        with Collection('. 781e5e245d69b566979b86e28d23f2c7+10 0:10:count1.txt') as c1:
            with Collection() as c2:
                f = c2.open("count1.txt", "wb")
                f.write("1234567890")

                self.assertTrue(c1["count1.txt"] != c2["count1.txt"])
                self.assertFalse(c1["count1.txt"] == c2["count1.txt"])


class ArvadosFileReadTestCase(unittest.TestCase, StreamRetryTestMixin):
    def reader_for(self, coll_name, **kwargs):
        stream = []
        segments = []
        n = 0
        for d in self.manifest_for(coll_name).split():
            try:
                k = KeepLocator(d)
                segments.append(Range(n, n, k.size))
                stream.append(Range(d, n, k.size))
                n += k.size
            except ValueError:
                pass

        blockmanager = arvados.arvfile._BlockManager(self.keep_client())
        blockmanager.prefetch_enabled = False
        col = Collection(keep_client=self.keep_client(), block_manager=blockmanager)
        af = ArvadosFile(col, "test",
                         stream=stream,
                         segments=segments)
        kwargs.setdefault('mode', 'rb')
        return ArvadosFileReader(af, **kwargs)

    def read_for_test(self, reader, byte_count, **kwargs):
        return reader.read(byte_count, **kwargs)


class ArvadosFileReadFromTestCase(ArvadosFileReadTestCase):
    def read_for_test(self, reader, byte_count, **kwargs):
        return reader.readfrom(0, byte_count, **kwargs)


class ArvadosFileReadAllTestCase(ArvadosFileReadTestCase):
    def read_for_test(self, reader, byte_count, **kwargs):
        return b''.join(reader.readall(**kwargs))


class ArvadosFileReadAllDecompressedTestCase(ArvadosFileReadTestCase):
    def read_for_test(self, reader, byte_count, **kwargs):
        return b''.join(reader.readall_decompressed(**kwargs))


class ArvadosFileReadlinesTestCase(ArvadosFileReadTestCase):
    def read_for_test(self, reader, byte_count, **kwargs):
        return ''.join(reader.readlines(**kwargs)).encode()


class ArvadosFileTestCase(unittest.TestCase):
    def datetime_to_hex(self, dt):
        return hex(int(time.mktime(dt.timetuple())))[2:]

    def test_permission_expired(self):
        base_manifest = ". 781e5e245d69b566979b86e28d23f2c7+10+A715fd31f8111894f717eb1003c1b0216799dd9ec@{} 0:10:count.txt\n"
        now = datetime.datetime.now()
        a_week_ago = now - datetime.timedelta(days=7)
        a_month_ago = now - datetime.timedelta(days=30)
        a_week_from_now = now + datetime.timedelta(days=7)
        with Collection(base_manifest.format(self.datetime_to_hex(a_week_from_now))) as c:
            self.assertFalse(c.find('count.txt').permission_expired())
        with Collection(base_manifest.format(self.datetime_to_hex(a_week_ago))) as c:
            f = c.find('count.txt')
            self.assertTrue(f.permission_expired())
            self.assertTrue(f.permission_expired(a_week_from_now))
            self.assertFalse(f.permission_expired(a_month_ago))


class BlockManagerTest(unittest.TestCase):
    def test_bufferblock_append(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        with arvados.arvfile._BlockManager(keep) as blockmanager:
            bufferblock = blockmanager.alloc_bufferblock()
            bufferblock.append("foo")

            self.assertEqual(bufferblock.size(), 3)
            self.assertEqual(bufferblock.buffer_view[0:3], b"foo")
            self.assertEqual(bufferblock.locator(), "acbd18db4cc2f85cedef654fccc4a4d8+3")

            bufferblock.append("bar")

            self.assertEqual(bufferblock.size(), 6)
            self.assertEqual(bufferblock.buffer_view[0:6], b"foobar")
            self.assertEqual(bufferblock.locator(), "3858f62230ac3c915f300c664312c63f+6")

            bufferblock.set_state(arvados.arvfile._BufferBlock.PENDING)
            with self.assertRaises(arvados.errors.AssertionError):
                bufferblock.append("bar")

    def test_bufferblock_dup(self):
        keep = ArvadosFileWriterTestCase.MockKeep({})
        with arvados.arvfile._BlockManager(keep) as blockmanager:
            bufferblock = blockmanager.alloc_bufferblock()
            bufferblock.append("foo")

            self.assertEqual(bufferblock.size(), 3)
            self.assertEqual(bufferblock.buffer_view[0:3], b"foo")
            self.assertEqual(bufferblock.locator(), "acbd18db4cc2f85cedef654fccc4a4d8+3")
            bufferblock.set_state(arvados.arvfile._BufferBlock.PENDING)

            bufferblock2 = blockmanager.dup_block(bufferblock, None)
            self.assertNotEqual(bufferblock.blockid, bufferblock2.blockid)

            bufferblock2.append("bar")

            self.assertEqual(bufferblock2.size(), 6)
            self.assertEqual(bufferblock2.buffer_view[0:6], b"foobar")
            self.assertEqual(bufferblock2.locator(), "3858f62230ac3c915f300c664312c63f+6")

            self.assertEqual(bufferblock.size(), 3)
            self.assertEqual(bufferblock.buffer_view[0:3], b"foo")
            self.assertEqual(bufferblock.locator(), "acbd18db4cc2f85cedef654fccc4a4d8+3")

    def test_bufferblock_get(self):
        keep = ArvadosFileWriterTestCase.MockKeep({
            "781e5e245d69b566979b86e28d23f2c7+10": b"0123456789",
        })
        with arvados.arvfile._BlockManager(keep) as blockmanager:
            bufferblock = blockmanager.alloc_bufferblock()
            bufferblock.append("foo")

            self.assertEqual(blockmanager.get_block_contents("781e5e245d69b566979b86e28d23f2c7+10", 1), b"0123456789")
            self.assertEqual(blockmanager.get_block_contents(bufferblock.blockid, 1), b"foo")

    def test_bufferblock_commit(self):
        mockkeep = mock.MagicMock()
        with arvados.arvfile._BlockManager(mockkeep) as blockmanager:
            bufferblock = blockmanager.alloc_bufferblock()
            bufferblock.owner = mock.MagicMock(spec=arvados.arvfile.ArvadosFile)
            def flush(sync=None):
                blockmanager.commit_bufferblock(bufferblock, sync)
            bufferblock.owner.flush.side_effect = flush
            bufferblock.append("foo")
            blockmanager.commit_all()
            self.assertTrue(bufferblock.owner.flush.called)
            self.assertTrue(mockkeep.put.called)
            self.assertEqual(bufferblock.state(), arvados.arvfile._BufferBlock.COMMITTED)
            self.assertIsNone(bufferblock.buffer_view)

    def test_bufferblock_commit_pending(self):
        # Test for bug #7225
        mockkeep = mock.MagicMock()
        mockkeep.put.side_effect = lambda *args, **kwargs: time.sleep(1)
        with arvados.arvfile._BlockManager(mockkeep) as blockmanager:
            bufferblock = blockmanager.alloc_bufferblock()
            bufferblock.append("foo")

            blockmanager.commit_bufferblock(bufferblock, False)
            self.assertEqual(bufferblock.state(), arvados.arvfile._BufferBlock.PENDING)

            blockmanager.commit_bufferblock(bufferblock, True)
            self.assertEqual(bufferblock.state(), arvados.arvfile._BufferBlock.COMMITTED)

    def test_bufferblock_commit_with_error(self):
        mockkeep = mock.MagicMock()
        mockkeep.put.side_effect = arvados.errors.KeepWriteError("fail")
        with arvados.arvfile._BlockManager(mockkeep) as blockmanager:
            bufferblock = blockmanager.alloc_bufferblock()
            bufferblock.owner = mock.MagicMock(spec=arvados.arvfile.ArvadosFile)
            def flush(sync=None):
                blockmanager.commit_bufferblock(bufferblock, sync)
            bufferblock.owner.flush.side_effect = flush
            bufferblock.append("foo")
            with self.assertRaises(arvados.errors.KeepWriteError) as err:
                blockmanager.commit_all()
            self.assertTrue(bufferblock.owner.flush.called)
            self.assertEqual(str(err.exception), "Error writing some blocks: block acbd18db4cc2f85cedef654fccc4a4d8+3 raised KeepWriteError (fail)")
            self.assertEqual(bufferblock.state(), arvados.arvfile._BufferBlock.ERROR)



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import next
from builtins import zip
from builtins import str
from builtins import range
import datetime
import itertools
import random
import unittest

from arvados.keep import KeepLocator

class ArvadosKeepLocatorTest(unittest.TestCase):
    DEFAULT_TEST_COUNT = 10

    def numstrs(fmtstr, base, exponent):
        def genstrs(self, count=None):
            return (fmtstr.format(random.randint(0, base ** exponent))
                    for c in range(count or self.DEFAULT_TEST_COUNT))
        return genstrs

    checksums = numstrs('{:032x}', 16, 32)
    sizes = numstrs('{:d}', 2, 26)
    signatures = numstrs('{:040x}', 16, 40)
    timestamps = numstrs('{:08x}', 16, 8)

    def base_locators(self, count=DEFAULT_TEST_COUNT):
        return ('+'.join(pair) for pair in
                zip(self.checksums(count), self.sizes(count)))

    def perm_hints(self, count=DEFAULT_TEST_COUNT):
        for sig, ts in zip(self.signatures(count),
                                      self.timestamps(count)):
            yield 'A{}@{}'.format(sig, ts)

    def test_good_locators_returned(self):
        for hint_gens in [(), (self.sizes(),),
                          (self.sizes(), self.perm_hints())]:
            for loc_data in zip(self.checksums(), *hint_gens):
                locator = '+'.join(loc_data)
                self.assertEqual(locator, str(KeepLocator(locator)))

    def test_nonchecksum_rejected(self):
        for badstr in ['', 'badbadbad', '8f9e68d957b504a29ba76c526c3145dj',
                       '+8f9e68d957b504a29ba76c526c3145d9',
                       '3+8f9e68d957b504a29ba76c526c3145d9']:
            self.assertRaises(ValueError, KeepLocator, badstr)

    def test_unknown_hints_accepted(self):
        base = next(self.base_locators(1))
        for weirdhint in ['Zfoo', 'Ybar234', 'Xa@b_c-372', 'W99']:
            locator = '+'.join([base, weirdhint])
            self.assertEqual(locator, str(KeepLocator(locator)))

    def test_bad_hints_rejected(self):
        base = next(self.base_locators(1))
        for badhint in ['', 'A', 'lowercase', '+32']:
            self.assertRaises(ValueError, KeepLocator,
                              '+'.join([base, badhint]))

    def test_multiple_locator_hints_accepted(self):
        base = next(self.base_locators(1))
        for loc_hints in itertools.permutations(['Kab1cd', 'Kef2gh', 'Kij3kl']):
            locator = '+'.join((base,) + loc_hints)
            self.assertEqual(locator, str(KeepLocator(locator)))

    def test_str_type(self):
        base = next(self.base_locators(1))
        locator = KeepLocator(base)
        self.assertEqual(type(''), type(locator.__str__()))

    def test_expiry_passed(self):
        base = next(self.base_locators(1))
        signature = next(self.signatures(1))
        dt1980 = datetime.datetime(1980, 1, 1)
        dt2000 = datetime.datetime(2000, 2, 2)
        dt2080 = datetime.datetime(2080, 3, 3)
        locator = KeepLocator(base)
        self.assertFalse(locator.permission_expired())
        self.assertFalse(locator.permission_expired(dt1980))
        self.assertFalse(locator.permission_expired(dt2080))
        # Timestamped to 1987-01-05 18:48:32.
        locator = KeepLocator('{}+A{}@20000000'.format(base, signature))
        self.assertTrue(locator.permission_expired())
        self.assertTrue(locator.permission_expired(dt2000))
        self.assertFalse(locator.permission_expired(dt1980))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
import os
import sys
import re

from setuptools import setup, find_packages

SETUP_DIR = os.path.dirname(__file__) or '.'
README = os.path.join(SETUP_DIR, 'README.rst')

import arvados_version
version = arvados_version.get_version(SETUP_DIR, "arvados")

short_tests_only = False
if '--short-tests-only' in sys.argv:
    short_tests_only = True
    sys.argv.remove('--short-tests-only')

setup(name='arvados-python-client',
      version=version,
      description='Arvados client library',
      long_description=open(README).read(),
      author='Arvados',
      author_email='info@arvados.org',
      url="https://arvados.org",
      download_url="https://github.com/arvados/arvados.git",
      license='Apache 2.0',
      packages=find_packages(),
      scripts=[
          'bin/arv-copy',
          'bin/arv-get',
          'bin/arv-keepdocker',
          'bin/arv-ls',
          'bin/arv-migrate-docker19',
          'bin/arv-federation-migrate',
          'bin/arv-normalize',
          'bin/arv-put',
          'bin/arv-ws'
      ],
      data_files=[
          ('share/doc/arvados-python-client', ['LICENSE-2.0.txt', 'README.rst']),
      ],
      install_requires=[
          'ciso8601 >=2.0.0',
          'future',
          'google-api-python-client >=1.6.2, <2',
          'google-auth<2',
          'httplib2 >=0.9.2, <0.20.2',
          'pycurl >=7.19.5.1, <7.45.0',
          'ruamel.yaml >=0.15.54, <0.17.22',
          'setuptools',
          'ws4py >=0.4.2',
          'protobuf<4.0.0dev',
          'pyparsing<3',
          'setuptools>=40.3.0',
      ],
      classifiers=[
          'Programming Language :: Python :: 3',
      ],
      test_suite='tests',
      tests_require=['pbr<1.7.0', 'mock>=1.0,<4', 'PyYAML'],
      zip_safe=False
      )



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import range
import collections
import http.client
import httplib2
import json
import logging
import os
import re
import socket
import ssl
import sys
import time
import types

import apiclient
from apiclient import discovery as apiclient_discovery
from apiclient import errors as apiclient_errors
from . import config
from . import errors
from . import util
from . import cache

_logger = logging.getLogger('arvados.api')

MAX_IDLE_CONNECTION_DURATION = 30
RETRY_DELAY_INITIAL = 2
RETRY_DELAY_BACKOFF = 2
RETRY_COUNT = 2

if sys.version_info >= (3,):
    httplib2.SSLHandshakeError = None

class OrderedJsonModel(apiclient.model.JsonModel):
    """Model class for JSON that preserves the contents' order.

    API clients that care about preserving the order of fields in API
    server responses can use this model to do so, like this::

        from arvados.api import OrderedJsonModel
        client = arvados.api('v1', ..., model=OrderedJsonModel())
    """

    def deserialize(self, content):
        # This is a very slightly modified version of the parent class'
        # implementation.  Copyright (c) 2010 Google.
        content = content.decode('utf-8')
        body = json.loads(content, object_pairs_hook=collections.OrderedDict)
        if self._data_wrapper and isinstance(body, dict) and 'data' in body:
            body = body['data']
        return body


def _intercept_http_request(self, uri, method="GET", headers={}, **kwargs):
    if not headers.get('X-Request-Id'):
        headers['X-Request-Id'] = self._request_id()
    try:
        if (self.max_request_size and
            kwargs.get('body') and
            self.max_request_size < len(kwargs['body'])):
            raise apiclient_errors.MediaUploadSizeError("Request size %i bytes exceeds published limit of %i bytes" % (len(kwargs['body']), self.max_request_size))

        if config.get("ARVADOS_EXTERNAL_CLIENT", "") == "true":
            headers['X-External-Client'] = '1'

        headers['Authorization'] = 'OAuth2 %s' % self.arvados_api_token

        retryable = method in [
            'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PUT']
        retry_count = self._retry_count if retryable else 0

        if (not retryable and
            time.time() - self._last_request_time > self._max_keepalive_idle):
            # High probability of failure due to connection atrophy. Make
            # sure this request [re]opens a new connection by closing and
            # forgetting all cached connections first.
            for conn in self.connections.values():
                conn.close()
            self.connections.clear()

        delay = self._retry_delay_initial
        for _ in range(retry_count):
            self._last_request_time = time.time()
            try:
                return self.orig_http_request(uri, method, headers=headers, **kwargs)
            except http.client.HTTPException:
                _logger.debug("[%s] Retrying API request in %d s after HTTP error",
                              headers['X-Request-Id'], delay, exc_info=True)
            except ssl.SSLCertVerificationError as e:
                raise ssl.SSLCertVerificationError(e.args[0], "Could not connect to %s\n%s\nPossible causes: remote SSL/TLS certificate expired, or was issued by an untrusted certificate authority." % (uri, e)) from None
            except socket.error:
                # This is the one case where httplib2 doesn't close the
                # underlying connection first.  Close all open
                # connections, expecting this object only has the one
                # connection to the API server.  This is safe because
                # httplib2 reopens connections when needed.
                _logger.debug("[%s] Retrying API request in %d s after socket error",
                              headers['X-Request-Id'], delay, exc_info=True)
                for conn in self.connections.values():
                    conn.close()

            time.sleep(delay)
            delay = delay * self._retry_delay_backoff

        self._last_request_time = time.time()
        return self.orig_http_request(uri, method, headers=headers, **kwargs)
    except Exception as e:
        # Prepend "[request_id] " to the error message, which we
        # assume is the first string argument passed to the exception
        # constructor.
        for i in range(len(e.args or ())):
            if type(e.args[i]) == type(""):
                e.args = e.args[:i] + ("[{}] {}".format(headers['X-Request-Id'], e.args[i]),) + e.args[i+1:]
                raise type(e)(*e.args)
        raise

def _patch_http_request(http, api_token):
    http.arvados_api_token = api_token
    http.max_request_size = 0
    http.orig_http_request = http.request
    http.request = types.MethodType(_intercept_http_request, http)
    http._last_request_time = 0
    http._max_keepalive_idle = MAX_IDLE_CONNECTION_DURATION
    http._retry_delay_initial = RETRY_DELAY_INITIAL
    http._retry_delay_backoff = RETRY_DELAY_BACKOFF
    http._retry_count = RETRY_COUNT
    http._request_id = util.new_request_id
    return http

def _close_connections(self):
    for conn in self._http.connections.values():
        conn.close()

# Monkey patch discovery._cast() so objects and arrays get serialized
# with json.dumps() instead of str().
_cast_orig = apiclient_discovery._cast
def _cast_objects_too(value, schema_type):
    global _cast_orig
    if (type(value) != type('') and
        type(value) != type(b'') and
        (schema_type == 'object' or schema_type == 'array')):
        return json.dumps(value)
    else:
        return _cast_orig(value, schema_type)
apiclient_discovery._cast = _cast_objects_too

# Convert apiclient's HttpErrors into our own API error subclass for better
# error reporting.
# Reassigning apiclient_errors.HttpError is not sufficient because most of the
# apiclient submodules import the class into their own namespace.
def _new_http_error(cls, *args, **kwargs):
    return super(apiclient_errors.HttpError, cls).__new__(
        errors.ApiError, *args, **kwargs)
apiclient_errors.HttpError.__new__ = staticmethod(_new_http_error)

def http_cache(data_type):
    homedir = os.environ.get('HOME')
    if not homedir or len(homedir) == 0:
        return None
    path = homedir + '/.cache/arvados/' + data_type
    try:
        util.mkdir_dash_p(path)
    except OSError:
        return None
    return cache.SafeHTTPCache(path, max_age=60*60*24*2)

def api(version=None, cache=True, host=None, token=None, insecure=False,
        request_id=None, timeout=5*60, **kwargs):
    """Return an apiclient Resources object for an Arvados instance.

    :version:
      A string naming the version of the Arvados API to use (for
      example, 'v1').

    :cache:
      Use a cache (~/.cache/arvados/discovery) for the discovery
      document.

    :host:
      The Arvados API server host (and optional :port) to connect to.

    :token:
      The authentication token to send with each API call.

    :insecure:
      If True, ignore SSL certificate validation errors.

    :timeout:
      A timeout value for http requests.

    :request_id:
      Default X-Request-Id header value for outgoing requests that
      don't already provide one. If None or omitted, generate a random
      ID. When retrying failed requests, the same ID is used on all
      attempts.

    Additional keyword arguments will be passed directly to
    `apiclient_discovery.build` if a new Resource object is created.
    If the `discoveryServiceUrl` or `http` keyword arguments are
    missing, this function will set default values for them, based on
    the current Arvados configuration settings.

    """

    if not version:
        version = 'v1'
        _logger.info("Using default API version. " +
                     "Call arvados.api('%s') instead." %
                     version)
    if 'discoveryServiceUrl' in kwargs:
        if host:
            raise ValueError("both discoveryServiceUrl and host provided")
        # Here we can't use a token from environment, config file,
        # etc. Those probably have nothing to do with the host
        # provided by the caller.
        if not token:
            raise ValueError("discoveryServiceUrl provided, but token missing")
    elif host and token:
        pass
    elif not host and not token:
        return api_from_config(
            version=version, cache=cache, timeout=timeout,
            request_id=request_id, **kwargs)
    else:
        # Caller provided one but not the other
        if not host:
            raise ValueError("token argument provided, but host missing.")
        else:
            raise ValueError("host argument provided, but token missing.")

    if host:
        # Caller wants us to build the discoveryServiceUrl
        kwargs['discoveryServiceUrl'] = (
            'https://%s/discovery/v1/apis/{api}/{apiVersion}/rest' % (host,))

    if 'http' not in kwargs:
        http_kwargs = {'ca_certs': util.ca_certs_path()}
        if cache:
            http_kwargs['cache'] = http_cache('discovery')
        if insecure:
            http_kwargs['disable_ssl_certificate_validation'] = True
        kwargs['http'] = httplib2.Http(**http_kwargs)

    if kwargs['http'].timeout is None:
        kwargs['http'].timeout = timeout

    kwargs['http'] = _patch_http_request(kwargs['http'], token)

    svc = apiclient_discovery.build('arvados', version, cache_discovery=False, **kwargs)
    svc.api_token = token
    svc.insecure = insecure
    svc.request_id = request_id
    svc.config = lambda: util.get_config_once(svc)
    svc.vocabulary = lambda: util.get_vocabulary_once(svc)
    svc.close_connections = types.MethodType(_close_connections, svc)
    kwargs['http'].max_request_size = svc._rootDesc.get('maxRequestSize', 0)
    kwargs['http'].cache = None
    kwargs['http']._request_id = lambda: svc.request_id or util.new_request_id()
    return svc

def api_from_config(version=None, apiconfig=None, **kwargs):
    """Return an apiclient Resources object enabling access to an Arvados server
    instance.

    :version:
      A string naming the version of the Arvados REST API to use (for
      example, 'v1').

    :apiconfig:
      If provided, this should be a dict-like object (must support the get()
      method) with entries for ARVADOS_API_HOST, ARVADOS_API_TOKEN, and
      optionally ARVADOS_API_HOST_INSECURE.  If not provided, use
      arvados.config (which gets these parameters from the environment by
      default.)

    Other keyword arguments such as `cache` will be passed along `api()`

    """
    # Load from user configuration or environment
    if apiconfig is None:
        apiconfig = config.settings()

    errors = []
    for x in ['ARVADOS_API_HOST', 'ARVADOS_API_TOKEN']:
        if x not in apiconfig:
            errors.append(x)
    if errors:
        raise ValueError(" and ".join(errors)+" not set.\nPlease set in %s or export environment variable." % config.default_config_file)
    host = apiconfig.get('ARVADOS_API_HOST')
    token = apiconfig.get('ARVADOS_API_TOKEN')
    insecure = config.flag_is_true('ARVADOS_API_HOST_INSECURE', apiconfig)

    return api(version=version, host=host, token=token, insecure=insecure, **kwargs)
